import json
import logging
import os
from dataclasses import dataclass
from datetime import datetime
from typing import List
import re
import sys

import numpy as np
import h5py
import pandas as pd
import torch
from genalog.text import alignment
from pandarallel import pandarallel
from tqdm import tqdm
from transformers import HfArgumentParser
import cohere
import replicate

from wtpsplit.evaluation import get_labels, evaluate_sentences_llm
from wtpsplit.evaluation.intrinsic_pairwise import generate_k_mers
from wtpsplit.utils import Constants
import time

pandarallel.initialize(progress_bar=True, nb_workers=32)

logging.getLogger().setLevel(logging.WARNING)

SYSTEM_PROMPT = (
    "Separate the following text into sentences by adding a newline between each sentence. "
    "Do not modify the text in any way and keep the exact ordering of words! "
    "If you modify it, remove or add anything, you get fined $1000 per word. "
    "Provide a concise answer without any introduction. "
    "Indicate sentence boundaries only via a single newline, no more than this! "
)

LYRICS_PROMPT = (
    "Separate the following song's lyrics into semantic units "
    "(e.g., verse, chorus, bridge, intro/outro, etc - "
    "similar to how they are presented in a lyrics booklet) "
    "via double newlines, but do not annotate them. "
    "Only include the song in the output, no annotations. "
    "Do not modify the song in any way and keep the exact ordering of words! "
    "If you modify it, remove or add anything, you get fined $1000 per word. "
    "Indicate semantic units by double newlines. "
)


@dataclass
class Args:
    eval_data_path: str = "data/all_data_11_05"
    type: str = "lyrics"  # all, lyrics, pairs, short_proc
    llm_provider: str = "cohere"  # cohere, replicate
    label_delimiter: str = "|"  # NOT \n or \n\n
    gap_char = "@"
    # model: str = "mistralai/mixtral-8x7b-instruct-v0.1"
    # model: str = "meta/meta-llama-3-8b-instruct"
    model: str = "command-r"
    save_suffix: str = "Pv2"
    include_langs: List[str] = None
    custom_language_list: str = None
    max_n_test_sentences: int = -1
    k: int = 10
    n_shots: int = 0


def replicate_provider(text, train_data, lang_code, args):
    api = replicate.Client(api_token=os.environ["REPLICATE_API_TOKEN"])
    llm_prompt = prompt_factory(text, train_data, lang_code, args)
    # print(llm_prompt)
    n_tries = 0
    while n_tries < 1:
        try:
            llm_input = {
                "system_prompt": "",
                "prompt": llm_prompt,
                # "max_new_tokens": 50_000,
                "max_tokens": 4000,
            }
            llm_output = api.run(args.model, llm_input)
            llm_output = "".join(llm_output)
            # print(llm_output)
            return llm_output
        except Exception as e:
            n_tries += 1
            print(e)
            time.sleep(10)
            continue
    return ""


def cohere_provider(text, train_data, lang_code, args):
    api = cohere.Client(os.environ["COHERE_API_TOKEN"])
    llm_prompt = prompt_factory(text, train_data, lang_code, args)
    n_tries = 0
    while True:
        try:
            llm_output = api.chat(
                model=args.model, preamble="", message=llm_prompt, max_tokens=4000, seed=42
            ).text.replace("\\n", "\n")
            return llm_output
        except Exception as e:
            status_code = getattr(e, "status_code", None)
            if status_code and str(e.status_code)[0] == "4":
                print("API refusal.")
                llm_output = ""
                return llm_output
            elif status_code and str(e.status_code)[0] == "5":
                # Cohere API issue: retry
                n_tries += 1
                time.sleep(10)
                continue
            else:
                raise e


def cohere_tokenize(text, args):
    api = cohere.Client(os.environ["COHERE_API_TOKEN"])
    tokens = api.tokenize(text=text, model=args.model, offline=False).token_strings
    return tokens


def load_h5_to_dataframe(filepath, args):
    with h5py.File(filepath, "r") as h5file:
        all_data = []

        for lang_code, lang_group in h5file.items():
            for dataset_name, dset_group in lang_group.items():
                if (
                    "test_preds" in dset_group
                    and "test_chunks" in dset_group
                    and not any("0" in key for key in dset_group.keys())
                ):
                    test_preds_data = dset_group["test_preds"].asstr()[:]
                    test_chunks_data = dset_group["test_chunks"].asstr()[:]

                    if len(test_preds_data) != len(test_chunks_data):
                        raise ValueError("Mismatched lengths between test_preds and test_chunks.")

                    # append  item with metadata
                    for test_pred, test_chunk in zip(test_preds_data, test_chunks_data):
                        all_data.append(
                            {
                                "lang": lang_code,
                                "dataset_name": dataset_name,
                                "test_preds": test_pred,
                                "test_chunks": test_chunk.tolist(),
                                "doc_id": -1,
                            }
                        )
                elif any("0" in key for key in dset_group.keys()):
                    # each list is saved with x_i, x_i+1, ...
                    doc_ids = [key.split("_")[-1] for key in dset_group.keys() if "test_preds" in key]
                    for doc_id in doc_ids:
                        test_preds_data = dset_group[f"test_preds_{doc_id}"].asstr()[:]
                        test_chunks_data = dset_group[f"test_chunks_{doc_id}"].asstr()[:]

                        if len(test_preds_data) != len(test_chunks_data):
                            raise ValueError("Mismatched lengths between test_preds and test_chunks.")

                        # append item with metadata
                        for test_pred, test_chunk in zip(test_preds_data, test_chunks_data):
                            all_data.append(
                                {
                                    "lang": lang_code,
                                    "dataset_name": dataset_name,
                                    "test_preds": test_pred,
                                    "test_chunks": test_chunk.tolist(),
                                    "doc_id": int(doc_id),
                                }
                            )
                else:
                    pass

        df = pd.DataFrame(all_data)
        return df


def load_or_compute_logits(args, eval_data, save_str: str = None):
    logits_dir = Constants.CACHE_DIR / "llm_sent" / "preds"
    logits_dir.mkdir(parents=True, exist_ok=True)
    logits_path = logits_dir / f"{save_str}.h5"

    if args.custom_language_list is not None:
        with open(args.custom_language_list, "r") as f:
            # file is a csv: l1,l2,...
            use_langs = f.read().strip().split(",")
    else:
        use_langs = eval_data.keys()

    with h5py.File(logits_path, "a") as f:
        for lang_code in tqdm(use_langs, desc="Languages"):
            if args.include_langs is not None and lang_code not in args.include_langs:
                continue

            print(f"Processing {lang_code}...")
            if lang_code not in f:
                lang_group = f.create_group(lang_code)
            else:
                lang_group = f[lang_code]
            for dataset_name, dataset in tqdm(eval_data[lang_code]["sentence"].items(), desc=lang_code):
                if "corrupted-asr" in dataset_name and (
                    "lyrics" not in dataset_name
                    and "short" not in dataset_name
                    and "code" not in dataset_name
                    and "ted" not in dataset_name
                    and "legal" not in dataset_name
                ):
                    print("SKIP: ", lang_code, dataset_name)
                    continue
                if "legal" in dataset_name and not ("laws" in dataset_name or "judgements" in dataset_name):
                    print("SKIP: ", lang_code, dataset_name)
                    continue
                if "social-media" in dataset_name:
                    continue
                if "nllb" in dataset_name:
                    continue
                if dataset_name not in lang_group:
                    dset_group = lang_group.create_group(dataset_name)
                if args.type == "pairs" and dataset_name != "ersatz" and dataset_name != "ted2020-corrupted-asr":
                    continue
                if (args.k != 10 or args.n_shots != 0) and dataset_name != "ersatz":
                    print("SKIP: ", lang_code, dataset_name)
                    continue
                else:
                    dset_group = lang_group[dataset_name]
                if "test_preds" not in dset_group and "test_preds_0" not in dset_group:
                    test_sentences = dataset["data"]
                    if not test_sentences:
                        continue
                    if (
                        isinstance(test_sentences[0], list)
                        and "lyrics" not in dataset_name
                        and "short" not in dataset_name
                        and args.type != "pairs"
                    ):
                        # documents: only 10% of documents. 1000 sentences --> 100 docs
                        max_n_sentences = args.max_n_test_sentences // 10
                        # shuffle docs
                        np.random.seed(42)
                        test_sentences = np.random.permutation(test_sentences).tolist()
                    else:
                        max_n_sentences = args.max_n_test_sentences
                    test_sentences = test_sentences[:max_n_sentences]
                    if isinstance(test_sentences[0], list) or args.type == "pairs":
                        if args.type == "pairs":
                            all_pairs = generate_k_mers(
                                test_sentences, k=2, do_lowercase=False, do_remove_punct=False, sample_pct=0.5
                            )
                            test_sentences = all_pairs
                        # list of lists: chunk each sublist
                        if "short" in dataset_name or "lyrics" in dataset_name or args.type == "pairs":
                            # only here: no chunking
                            test_chunks = test_sentences
                            test_texts = [
                                [Constants.SEPARATORS[lang_code].join(test_chunk).strip()] for test_chunk in test_chunks
                            ]
                        else:
                            test_chunks = [
                                [test_sentences[i][j : j + args.k] for j in range(0, len(test_sentences[i]), args.k)]
                                for i in range(len(test_sentences))
                            ]
                            # last batch for each sublist: pad with None to enable saving w/ h5py
                            for i in range(len(test_chunks)):
                                test_chunks[i][-1] += [""] * (args.k - len(test_chunks[i][-1]))
                            # join sentences in each chunk
                            test_texts = [
                                [
                                    Constants.SEPARATORS[lang_code].join(test_chunk).strip()
                                    for test_chunk in test_sublist
                                ]
                                for test_sublist in test_chunks
                            ]
                        if args.n_shots:
                            train_sentences = eval_data[lang_code]["sentence"][dataset_name]["meta"]["train_data"][:100]
                            if train_sentences:
                                if "short" in dataset_name or args.type == "pairs":
                                    # here: entire samples (tweets e.g.)
                                    train_chunks = train_sentences
                                    train_texts = ["\n".join(train_chunk).strip() for train_chunk in train_chunks]
                                elif "lyrics" in dataset_name:
                                    # here: entire samples (songs)
                                    train_chunks = train_sentences
                                    train_texts = ["\n\n".join(train_chunk).strip() for train_chunk in train_chunks]
                                else:
                                    # flatten to have diversity among lengthy few-shot examples
                                    train_sentences = [item for sublist in train_sentences[:100] for item in sublist]
                                    train_chunks = [
                                        train_sentences[i : i + args.k] for i in range(0, len(train_sentences), args.k)
                                    ]
                                    train_texts = ["\n".join(train_chunk).strip() for train_chunk in train_chunks]
                            else:
                                train_texts = None
                        else:
                            train_texts = None
                        for i in tqdm(range(len(test_texts))):
                            test_preds = get_llm_preds(test_texts[i], train_texts, lang_code, args, verbose=False)
                            dset_group.create_dataset(f"test_preds_{i}", data=test_preds)
                            dset_group.create_dataset(
                                f"test_chunks_{i}",
                                data=[test_chunks[i]]
                                if "short" in dataset_name or "lyrics" in dataset_name or args.type == "pairs"
                                else test_chunks[i],
                            )

                    else:
                        test_chunks = [test_sentences[i : i + args.k] for i in range(0, len(test_sentences), args.k)]
                        # last batch: pad with None to enable saving w/ h5py
                        test_chunks[-1] += [""] * (args.k - len(test_chunks[-1]))
                        test_texts = [
                            Constants.SEPARATORS[lang_code].join(test_chunk).strip() for test_chunk in test_chunks
                        ]
                        if args.n_shots:
                            train_sentences = eval_data[lang_code]["sentence"][dataset_name]["meta"]["train_data"]
                            if train_sentences:
                                train_sentences = train_sentences[:100]
                                train_chunks = [
                                    train_sentences[i : i + args.k] for i in range(0, len(train_sentences), args.k)
                                ]
                                train_texts = ["\n".join(train_chunk).strip() for train_chunk in train_chunks]
                            else:
                                train_texts = None
                        else:
                            train_texts = None
                        test_preds = get_llm_preds(test_texts, train_texts, lang_code, args)
                        dset_group.create_dataset("test_preds", data=test_preds)
                        dset_group.create_dataset("test_chunks", data=test_chunks)
    return h5py.File(logits_path, "r")


def get_llm_preds(
    test_texts,
    train_data,
    lang_code,
    args,
    verbose=True,
):
    if args.llm_provider == "cohere":
        llm_provider = cohere_provider
    elif args.llm_provider == "replicate":
        llm_provider = replicate_provider
    else:
        raise ValueError(f"Unknown LLM provider: {args.llm_provider}")
    output = []
    for test_chunk in tqdm(test_texts, disable=not verbose):
        output.append(llm_provider(test_chunk, train_data, lang_code, args))
    return output


def prompt_factory(test_chunk, train_data, lang_code, args):
    n_shots = args.n_shots if train_data is not None else 0
    main_prompt = LYRICS_PROMPT if args.type == "lyrics" else SYSTEM_PROMPT

    prompt_start = (
        main_prompt
        + f"When provided with multiple examples, you are to respond only to the last one: Output {n_shots + 1}."
        if n_shots
        else main_prompt
    )

    llm_prompt = (
        prompt_start
        + "\n\n"
        + create_few_shot_prompt(train_data, lang_code, args)
        + (f"# Input {n_shots + 1}:\n\n" if n_shots else "# Input:\n\n")
        + test_chunk
        + (f"\n\n# Output {n_shots + 1}:\n\n" if n_shots else "\n\n# Output:\n\n")
    )
    return llm_prompt


def create_few_shot_prompt(train_data, lang_code, args):
    if train_data is None:
        return ""
    num_samples = min(args.n_shots, len(train_data))
    samples_prompt = ""
    counter = 1
    for _, sample in enumerate(train_data[:num_samples]):
        current_input = sample.replace("\n", Constants.SEPARATORS.get(lang_code, " "))
        samples_prompt += f"# Input {counter}:\n\n{current_input}\n\n# Output {counter}:\n\n{sample}\n\n"
        counter += 1
    return samples_prompt


def postprocess_llm_output(llm_output, lang):
    """Clean LLM output by removing specified characters and cleaning up lines."""
    if llm_output == "":
        # API refusal
        return ""
    llm_output = llm_output.strip(" -\n")

    # neither of them must be present in the output.
    llm_output = llm_output.replace(args.gap_char, " ")
    llm_output = llm_output.replace(args.label_delimiter, " ")
    llm_output = llm_output.replace("\n\n", args.label_delimiter)
    llm_output = llm_output.replace("\n", args.label_delimiter)
    # replace multiple newlines with 1
    llm_output = re.sub(r"\n+", "\n", llm_output)

    # remove leading #, # Input, :
    llm_output = llm_output.strip("#").strip().strip("Input").strip(":").strip()
    # remove trailing #, Output, .
    llm_output = llm_output.strip(":").strip("Output").strip().strip("#")
    # replace multiple occurences of label_delimiter with only 1
    llm_output = re.sub(r"{0}+".format(re.escape(args.label_delimiter)), args.label_delimiter, llm_output)

    # Split into lines, strip each line, remove empty lines, and join back into a single string
    llm_output = " ".join([line.strip() for line in llm_output.split("\n") if line.strip()])

    return llm_output


def align_llm_output(row):
    """Align input and output, including formatting."""
    try:
        aligned_in, aligned_llm = alignment.align(
            row["test_chunks"],
            row["test_preds"],
            gap_char=args.gap_char,
        )
        # same as aligned_in, aligned_llm, but with additional formatting. Latter used to debug only.
        formatted_alignment = alignment._format_alignment(aligned_in, aligned_llm).split("\n")
    except:  # noqa
        print("Alignment failed: ", row.name)
        formatted_alignment = [row["test_chunks"], "", " " * len(row["test_chunks"])]
    return pd.Series(
        {
            "alignment": formatted_alignment[1],
            "alignment_in": formatted_alignment[0],
            "alignment_out": formatted_alignment[2],
        }
    )


def process_alignment(row, args):
    gt = row["alignment_in"]
    preds = row["alignment_out"]
    old_label_count = preds.count("|")

    # we need the following to ensure that no label is overwritten.
    processed_gt = []
    processed_preds = []

    pred_index = 0  # separate index for preds to handle skipped characters in gt

    for gt_char in gt:
        if gt_char == args.gap_char:
            # check if the corresponding preds char is a label delimiter and needs to be shifted
            if pred_index < len(preds) and preds[pred_index] == args.label_delimiter:
                # only shift if there's room and it's safe to shift back
                if processed_preds:
                    processed_preds[-1] = args.label_delimiter
            # do not add the gap character from gt to processed_gt
        else:
            processed_gt.append(gt_char)
            # only add characters from preds if not out of bounds
            if pred_index < len(preds):
                processed_preds.append(preds[pred_index])

        # Increment regardless of whether it's a gap to keep alignment
        pred_index += 1

    gt = "".join(processed_gt)
    preds = "".join(processed_preds)
    # re-check
    new_label_count = preds.count(args.label_delimiter)
    if old_label_count != new_label_count:
        # happens in some garbage LLM predictions, which can be easily picked up, so it is fine (favors LLM; less FPs)
        print("Gap char not removed correctly or labels shifted improperly: ", row.name, row.lang, row.dataset_name)
    assert len(gt) == len(preds), f"Length mismatch: {len(gt)} vs {len(preds)}"

    if Constants.SEPARATORS.get(row["lang"], " ") == "":
        # ensure that, after label calculation, both seqs are of same lengths & labels at proper positions.
        missing_indices = [
            i for i, char in enumerate(gt) if char == args.label_delimiter and preds[i] != args.label_delimiter
        ]
        extra_indices = [
            i for i, char in enumerate(preds) if char == args.label_delimiter and gt[i] != args.label_delimiter
        ]
        preds = "".join([char for i, char in enumerate(preds) if i not in missing_indices])
        for i in extra_indices:
            preds = preds[:i] + " " + preds[i:]

    # GT
    sentences = gt.split(args.label_delimiter)
    labels = get_labels(row.lang, sentences)
    # PREDS
    pred_sentences = preds.split(args.label_delimiter)
    predictions = get_labels(row.lang, pred_sentences)
    return labels, predictions


def calc_hallucination_deletion_rate(row):
    gt = row["alignment_in"]
    preds = row["alignment_out"]
    if all([char == args.gap_char for char in preds]):
        # all @: alignment failure, just garbage output
        return 0.0, 0.0

    hallucination_count = 0
    deletion_count = 0

    for gt_char, pred_char in zip(gt, preds):
        if gt_char == args.gap_char and pred_char != args.label_delimiter:
            hallucination_count += 1
        if pred_char == args.gap_char and gt_char != args.label_delimiter:
            deletion_count += 1

    deletion_rate = deletion_count / len(gt)
    hallucination_rate = hallucination_count / len(gt)
    return hallucination_rate, deletion_rate


def calculate_global_metric_averages(results):
    """dict of results[lang_code][dataset_name] -> dict of metrics -> float"""
    metric_totals = {}
    metric_counts = {}

    for lang_datasets in results.values():
        for metrics in lang_datasets.values():
            # aggregate
            metrics = metrics[args.model]
            for metric_key, metric_value in metrics.items():
                if isinstance(metric_value, (int, float)):
                    if metric_key not in metric_totals:
                        metric_totals[metric_key] = 0
                        metric_counts[metric_key] = 0

                    metric_totals[metric_key] += metric_value
                    metric_counts[metric_key] += 1

    # global average
    global_metric_averages = {
        metric: metric_totals[metric] / metric_counts[metric] for metric in metric_totals if metric_counts[metric] > 0
    }

    return global_metric_averages


def main(args):
    print(args)

    # Define paths for different categories
    avg_dir = Constants.CACHE_DIR / "llm_sent" / "AVG"
    default_dir = Constants.CACHE_DIR / "llm_sent" / "results"
    alignment_dir = Constants.CACHE_DIR / "llm_sent" / "alignments"

    # Ensure directories exist
    avg_dir.mkdir(parents=True, exist_ok=True)
    default_dir.mkdir(parents=True, exist_ok=True)
    alignment_dir.mkdir(parents=True, exist_ok=True)

    if args.type == "all" or args.type == "pairs":
        eval_data_path = args.eval_data_path + "-all.pth"
    elif args.type == "lyrics":
        eval_data_path = args.eval_data_path + "-lyrics.pth"
    elif args.type == "short_proc":
        eval_data_path = args.eval_data_path + "-short_proc.pth"
    else:
        raise ValueError(f"Unknown type: {args.type}")

    assert len(args.gap_char) == len(args.label_delimiter) == 1

    eval_data = torch.load(eval_data_path)

    save_str = (f"{args.model.split('/')[-1]}_k{args.k}_s{args.n_shots}").replace("/", "_")

    if args.max_n_test_sentences < sys.maxsize and args.max_n_test_sentences != -1:
        save_str += f"_n{args.max_n_test_sentences}"
    if args.max_n_test_sentences == -1:
        args.max_n_test_sentences = sys.maxsize
    save_str += f"{args.save_suffix}"
    save_str += f"-{args.type}"

    print(save_str)

    outputs = load_or_compute_logits(args, eval_data, save_str)

    # create df based on test_chunks and test_logits
    df = load_h5_to_dataframe(outputs.filename, args)
    print("Loaded df.")
    # df = df[df["lang"].isin(["ja", "en"])]  # DEBUG

    # postprocess
    df["test_preds"] = df.apply(lambda row: postprocess_llm_output(row["test_preds"], row["lang"]), axis=1)

    # remove empty strings (just needed for h5py storage)
    df["test_chunks"] = df["test_chunks"].apply(lambda x: [item for item in x if item != ""])
    # replace \n
    # NOTE: length and labels remain the same, crucially!
    df["test_chunks"] = df.apply(
        lambda row: args.label_delimiter.join(
            chunk.replace(args.label_delimiter, " ").replace(args.gap_char, " ") for chunk in row["test_chunks"]
        ),
        axis=1,
    )
    print("Processed df.")

    # align with Needleman Wunsch algorithm
    alignment = df.parallel_apply(align_llm_output, axis=1)
    df = df.join(alignment)
    print("Aligned df.")

    def concatenate_texts(group):
        # refusal: @@@@ --> taken into account with success_METRIC
        # alignment failure: "     " --> garbage output, no pos. labels
        return pd.Series(
            {
                "test_preds": args.label_delimiter.join(group["test_preds"]),
                "test_chunks": args.label_delimiter.join(group["test_chunks"]),
                "alignment": args.label_delimiter.join(group["alignment"]),
                "alignment_in": args.label_delimiter.join(group["alignment_in"]),
                "alignment_out": args.label_delimiter.join(group["alignment_out"]),
            }
        )

    # concat chunks
    # old_df = df.copy()
    df = df.groupby(["lang", "dataset_name", "doc_id"]).apply(concatenate_texts).reset_index()

    df["hallucination_rate"], df["deletion_rate"] = zip(
        *df.apply(
            calc_hallucination_deletion_rate,
            axis=1,
        )
    )

    results = {}
    indices = {}
    for lang_code in df["lang"].unique():
        results[lang_code] = {}
        indices[lang_code] = {}
        for dataset_name in df["dataset_name"].unique():
            results[lang_code][dataset_name] = {args.model: {}}  # Initialize nested dict with model
            indices[lang_code][dataset_name] = {args.model: {}}
            if "lyrics" in dataset_name or "short" in dataset_name or args.type == "pairs":
                exclude_every_k = 0
            else:
                exclude_every_k = args.k
            n_docs = len(df[(df["lang"] == lang_code) & (df["dataset_name"] == dataset_name)])
            if n_docs == 0:
                # combination non-existing
                continue
            indices[lang_code][dataset_name][args.model] = {}
            if n_docs > 1:
                # list of lists, TODO
                rows = df[(df["lang"] == lang_code) & (df["dataset_name"] == dataset_name)]
                metrics = []
                for i, row in rows.iterrows():
                    # apply processing before label calculation
                    labels, preds = process_alignment(row, args)
                    doc_metrics = {}
                    doc_metrics["refused"] = [
                        int(len(set(row["alignment_out"])) == 1 and args.gap_char in set(row["alignment_out"]))
                    ]
                    if doc_metrics["refused"][0]:
                        preds[-1] = 0
                    doc_metrics.update(
                        evaluate_sentences_llm(labels, preds, return_indices=True, exclude_every_k=exclude_every_k)
                    )
                    doc_metrics["length"] = [doc_metrics["length"]]
                    doc_metrics["hallucination_rate"] = row["hallucination_rate"]
                    doc_metrics["deletion_rate"] = row["deletion_rate"]

                    metrics.append(doc_metrics)
                # Initialization and collection of data
                avg_results = {}
                concat_indices = {}
                for doc in metrics:
                    for key, value in doc.items():
                        if isinstance(value, (float, int)):
                            # Store all numeric results
                            if key not in avg_results:
                                avg_results[key] = []
                                avg_results[key + "_success"] = []  # Initialize success list

                            # Append to the general results
                            avg_results[key].append(value)

                            # Append to success results if not refused
                            if not doc["refused"][0]:
                                avg_results[key + "_success"].append(value)
                        elif isinstance(value, list):
                            # Concatenate list values, handle 'refused' by only adding the first item of the list
                            if key not in concat_indices:
                                concat_indices[key] = []
                            if key == "refused" or key == "length":
                                concat_indices[key].append(value[0])  # Only the first item for 'refused'
                            else:
                                concat_indices[key].append(value)  # Extend with the full list otherwise

                # Calculate the average for numeric values and success metrics
                for key in list(avg_results):  # Use list to include newly added success keys safely during iteration
                    if avg_results[key]:  # Ensure there's data to calculate average
                        avg_results[key] = sum(avg_results[key]) / len(avg_results[key])

                # Store the results and indices
                results[lang_code][dataset_name][args.model] = avg_results
                indices[lang_code][dataset_name][args.model] = concat_indices
            else:
                # one long string
                row = df[(df["lang"] == lang_code) & (df["dataset_name"] == dataset_name)].iloc[0]

                # apply processing before label calculation
                labels, preds = process_alignment(row, args)
                # metrics!
                metrics = evaluate_sentences_llm(labels, preds, return_indices=True, exclude_every_k=exclude_every_k)
                metrics["hallucination_rate"] = row["hallucination_rate"]
                metrics["deletion_rate"] = row["deletion_rate"]
                indices[lang_code][dataset_name][args.model]["true_indices"] = [metrics.pop("true_indices")]
                indices[lang_code][dataset_name][args.model]["predicted_indices"] = [metrics.pop("predicted_indices")]
                indices[lang_code][dataset_name][args.model]["length"] = [metrics.pop("length")]
                results[lang_code][dataset_name][args.model] = metrics

    out_dict = {
        "metrics": calculate_global_metric_averages(results),
        "include_langs": args.include_langs,
        "max_n_test_sentences": args.max_n_test_sentences,
        "k": args.k,
        "n_success": len(df[df["test_preds"] != ""]),
        "success_rate": len(df[df["test_preds"] != ""]) / len(df),
        "model": args.model,
        "time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        "system_prompt": LYRICS_PROMPT if args.type == "lyrics" else SYSTEM_PROMPT,
    }

    json.dump(
        out_dict,
        open(
            avg_dir / f"{save_str}.json",
            "w",
            encoding="utf-8",
        ),
        indent=4,
    )
    json.dump(
        results,
        open(
            default_dir / f"{save_str}.json",
            "w",
            encoding="utf-8",
        ),
        indent=4,
    )
    json.dump(
        indices,
        open(
            alignment_dir / f"{save_str}_IDX.json",
            "w",
        ),
        default=int,
        indent=4,
    )
    print(alignment_dir / f"{save_str}_IDX.json")
    print("Indices saved to file.")

    df.to_csv(default_dir / "csv" / f"{save_str}.csv", index=False)
    print(out_dict)
    print(save_str)

    print(avg_dir / f"{save_str}.json")
    print(default_dir / f"{save_str}.json")


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()
    main(args)



import glob
import os
from dataclasses import dataclass

import conllu
import numpy as np
import torch
from datasets import load_dataset
from tqdm.auto import tqdm
from transformers import HfArgumentParser

from wtpsplit.evaluation import preprocess_sentence
from wtpsplit.utils import Constants

UD_TREEBANK_PATH = "data/external/ud-treebanks-v2.10"  # source: https://universaldependencies.org/#download
WIKTEXTRACT_DATA_PATH = "data/external/raw-wiktextract-data.json"  # source: https://kaikki.org/dictionary/rawdata.html
# source:
# https://uni-tuebingen.de/fakultaeten/philosophische-fakultaet/fachbereiche/neuphilologie/seminar-fuer-sprachwissenschaft/arbeitsbereiche/allg-sprachwissenschaft-computerlinguistik/ressourcen/lexica/germanet-1/beschreibung/compounds/
GERMANET_DATA_PATH = "data/external/split_compounds_from_GermaNet17.0_modified-2022-06-28.txt"
AUCOPRO_DATA_PATH = "data/external/Data.AUCOPRO.Splitting"  # source: https://sourceforge.net/projects/aucopro/
ERSATZ_DATA_PATH = "data/external/ersatz-test-suite/segmented"  # source: https://github.com/rewicks/ersatz-test-suite

# copied from Table 8 in https://aclanthology.org/2021.acl-long.309.pdf
ERSATZ_TEST_DATASETS = {
    "ar": "iwsltt2017.ar",
    "cs": "wmt20.cs-en.cs",
    "de": "wmt20.de-en.de",
    "en": "wsj.03-06.en",
    "es": "wmt13.es-en.es",
    "et": "wmt18.et-en.et",
    "fi": "wmt19.fi-en.fi",
    "fr": "wmt20.fr-de.fr",
    "gu": "wmt19.gu-en.gu",
    "hi": "wmt14.hi-en.hi",
    "iu": "wmt20.iu-en.iu",
    "ja": "wmt20.ja-en.ja",
    "kk": "wmt19.kk-en.kk",
    "km": "wmt20.km-en.km",
    "lt": "wmt19.lt-en.lt",
    "lv": "wmt17.lv-en.lv",
    "pl": "wmt20.pl-en.pl",
    "ps": "wmt20.ps-en.ps",
    "ro": "wmt16.ro-en.ro",
    "ru": "wmt20.ru-en.ru",
    "ta": "wmt20.ta-en.ta",
    "tr": "wmt18.tr-en.tr",
    "zh": "wmt20.zh-en.zh",
}
ERSATZ_TRAIN_DATASETS = {
    "ar": "news-commentary-v15.dev.ar",
    "cs": "wmt18.cs-en.cs",
    "de": "wmt19.de-en.de",
    "en": "merged.nc-wsj.en",
    "es": None,
    "et": "newscrawl.2019.dev.et",
    "fi": "wmt18.fi-en.fi",
    "fr": "wmt15.fr-en.fr",
    "gu": "newscrawl.2019.dev.gu",
    "hi": "newscrawl.2013.dev.hi",
    "iu": "nhi-3.0.iu",
    "ja": "newscrawl.2019.dev.ja",
    "kk": "newscrawl.2019.dev.kk",
    "km": "wikimatrix.dev.km",
    "lt": "newscrawl.2019.dev.lt",
    "lv": "newscrawl.2019.dev.lv",
    "pl": "newscrawl.2019.dev.pl",
    "ps": "wikimatrix.dev.ps",
    "ro": "newscrawl.2019.dev.ro",
    "ru": "wmt18.ru-en.ru",
    "ta": "newscrawl.2019.dev.ta",
    "tr": "wmt16.tr-en.tr",
    "zh": "wmt18.zh-en.zh",
}


@dataclass
class Args:
    output_file: str = "data/eval.pth"
    include_train_data: bool = False


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()

    eval_data = {lang_code: {"sentence": {}, "compound": {}} for lang_code in Constants.LANGINFO.index}

    # Ersatz data
    for lang_code in tqdm(Constants.LANGINFO.index):
        if lang_code in ERSATZ_TEST_DATASETS:
            eval_data[lang_code]["sentence"]["ersatz"] = {
                "meta": {
                    "train_data": [
                        preprocess_sentence(line)
                        for line in open(
                            os.path.join(
                                ERSATZ_DATA_PATH,
                                lang_code,
                                ERSATZ_TRAIN_DATASETS[lang_code],
                            )
                        )
                    ]
                    if args.include_train_data and ERSATZ_TRAIN_DATASETS[lang_code] is not None
                    else None,
                },
                "data": [
                    preprocess_sentence(line)
                    for line in open(os.path.join(ERSATZ_DATA_PATH, lang_code, ERSATZ_TEST_DATASETS[lang_code]))
                ],
            }

    # UD + OPUS100 sentences
    for lang_code in tqdm(Constants.LANGINFO.index):
        opus_dset_name = Constants.LANGINFO.loc[lang_code, "opus100"]

        if opus_dset_name not in (np.nan, None):
            other_lang_code = set(opus_dset_name.split("-")) - {lang_code}
            assert len(other_lang_code) == 1
            other_lang_code = other_lang_code.pop()

            dset_args = ["opus100", opus_dset_name]

            try:
                opus100_sentences = [
                    preprocess_sentence(sample[lang_code])
                    for sample in load_dataset(*dset_args, split="test")["translation"]
                    if sample[lang_code].strip() != sample[other_lang_code].strip()
                ]
                try:
                    opus100_train_sentences = [
                        preprocess_sentence(sample[lang_code])
                        for sample in load_dataset(*dset_args, split="train")["translation"]
                        if sample[lang_code].strip() != sample[other_lang_code].strip()
                    ]
                except ValueError:
                    opus100_train_sentences = None
            except ValueError:
                opus100_sentences = [
                    preprocess_sentence(sample[lang_code])
                    for sample in load_dataset(*dset_args, split="train")["translation"]
                    if sample[lang_code].strip() != sample[other_lang_code].strip()
                ]
                opus100_train_sentences = None

            eval_data[lang_code]["sentence"]["opus100"] = {
                "meta": {"train_data": opus100_train_sentences if args.include_train_data else None},
                "data": opus100_sentences,
            }

        if Constants.LANGINFO.loc[lang_code, "ud"] not in (np.nan, None):
            ud_data = conllu.parse(
                open(
                    glob.glob(
                        os.path.join(
                            UD_TREEBANK_PATH,
                            Constants.LANGINFO.loc[lang_code, "ud"],
                            "*-ud-test.conllu",
                        )
                    )[0]
                ).read()
            )

            try:
                ud_train_data = conllu.parse(
                    open(
                        glob.glob(
                            os.path.join(
                                UD_TREEBANK_PATH,
                                Constants.LANGINFO.loc[lang_code, "ud"],
                                "*-ud-train.conllu",
                            )
                        )[0]
                    ).read()
                )
                ud_train_sentences = [preprocess_sentence(sentence.metadata["text"]) for sentence in ud_train_data]
            except IndexError:
                ud_train_sentences = None

            ud_sentences = [preprocess_sentence(sentence.metadata["text"]) for sentence in ud_data]
            eval_data[lang_code]["sentence"]["ud"] = {
                "meta": {"train_data": ud_train_sentences if args.include_train_data else None},
                "data": ud_sentences,
            }

    torch.save(eval_data, args.output_file)



import json
import logging
import os
import sys
import time
from dataclasses import dataclass
from typing import List
import warnings

import h5py
import numpy as np
import torch
from tqdm.auto import tqdm
from transformers import AutoModelForTokenClassification, AutoTokenizer, HfArgumentParser, pipeline

from wtpsplit.evaluation import evaluate_mixture
from wtpsplit.utils import Constants

logger = logging.getLogger()
logger.setLevel(logging.INFO)

warnings.filterwarnings("ignore", category=UserWarning)


@dataclass
class Args:
    eval_data_path: str = "data/all_data_11_05-all.pth"
    device: str = "cpu"
    include_langs: List[str] = None
    max_n_test_sentences: int = sys.maxsize
    stride: int = 64
    save_suffix: str = ""
    return_indices: bool = False
    type: str = "both"  # laws, judgements, both, specific
    lang_support: str = "multi"  # mono, multi


def get_law_preds(texts, model, model_name, args) -> List[List[int]]:
    tokenizer = AutoTokenizer.from_pretrained(model_name)

    pipe = pipeline(
        "token-classification",
        model=model,
        tokenizer=tokenizer,
        device=args.device,
        aggregation_strategy="simple",
        stride=args.stride,
    )
    sentences = pipe(texts)
    sent_end_preds_all = []
    for text, sent_preds in zip(texts, sentences):
        sent_end_indices = [short_seq["end"] - 1 for short_seq in sent_preds]
        # indices to binary list
        sent_end_preds = [1 if i in sent_end_indices else 0 for i in range(len(text))]
        sent_end_preds_all.append(sent_end_preds)
    return sent_end_preds_all


def load_or_compute_logits(args, eval_data, save_str: str = None):
    logits_path = Constants.CACHE_DIR / "law_bert" / f"{save_str}.h5"
    base_name = "rcds/distilbert-SBD"  # take from HF hub

    if not os.path.exists(Constants.CACHE_DIR / "law_bert"):
        os.makedirs(Constants.CACHE_DIR / "law_bert")

    use_langs = ["fr", "es", "it", "en", "de", "pt"]

    total_test_time = 0

    with h5py.File(logits_path, "a") as f, torch.no_grad():
        for lang_code in tqdm(use_langs, desc="Languages"):
            if args.include_langs is not None and lang_code not in args.include_langs:
                continue
            if lang_code not in f:
                lang_group = f.create_group(lang_code)
            else:
                lang_group = f[lang_code]

            # eval data
            for dataset_name, dataset in tqdm(eval_data[lang_code]["sentence"].items(), desc=lang_code):
                if "legal" not in dataset_name:
                    continue
                if "legal" in dataset_name and not ("laws" in dataset_name or "judgements" in dataset_name):
                    continue
                if "social-media" in dataset_name:
                    continue
                current_name = base_name
                # map to correct model
                if args.lang_support == "multi":
                    current_name += "-fr-es-it-en-de"
                elif args.lang_support == "mono":
                    if lang_code.split("_")[0] == "pt":
                        current_name += "-fr-es-it-en-de"
                    else:
                        current_name += f"-{lang_code.split('_')[0]}"
                    if lang_code.split("_")[0] == "en":
                        current_name += "-judgements-laws"
                else:
                    raise NotImplementedError
                if lang_code.split("_")[0] == "en" and args.lang_support == "mono":
                    pass
                elif args.type == "laws":
                    current_name += "-laws"
                elif args.type == "judgements":
                    current_name += "-judgements"
                elif args.type == "both":
                    current_name += "-judgements-laws"
                elif args.type == "specific":
                    current_name += f"-{dataset_name.split('-')[-1]}"
                else:
                    raise NotImplementedError

                model = AutoModelForTokenClassification.from_pretrained(current_name).to(args.device)
                logger.info(f"RUN {lang_code} {dataset_name} {current_name}")
                if dataset_name not in lang_group:
                    dset_group = lang_group.create_group(dataset_name)
                else:
                    dset_group = lang_group[dataset_name]

                if "test_logits" not in dset_group:
                    test_sentences = dataset["data"][: args.max_n_test_sentences]
                    if not test_sentences:
                        continue
                    if isinstance(test_sentences[0], list):
                        # short-seq eval: list of lists
                        test_text = [
                            Constants.SEPARATORS.get(lang_code, " ").join(sentence) for sentence in test_sentences
                        ]
                    else:
                        raise NotImplementedError

                    start_time = time.time()
                    test_logits = get_law_preds(test_text, model, current_name, args)
                    end_time = time.time()
                    total_test_time += end_time - start_time
                    if isinstance(test_sentences[0], list):
                        test_logit_lengths = []
                        # store start and end indices for each pair, used later to slice the logits
                        # (h5py does not like different length np arrays as list elements)
                        all_logit_lengths = np.append(0, np.cumsum([len(logits) for logits in test_logits]))
                        # append tuple of start and end indices for each pair
                        for i in range(len(test_logits)):
                            test_logit_lengths.append((all_logit_lengths[i], all_logit_lengths[i + 1] - 1))
                        test_logits = np.concatenate(test_logits)
                        test_logits = np.expand_dims(test_logits, axis=1)
                        dset_group.create_dataset("test_logit_lengths", data=test_logit_lengths)
                    else:
                        raise NotImplementedError

                    dset_group.create_dataset("test_logits", data=test_logits)

    end_time = time.time()
    return h5py.File(logits_path, "r"), total_test_time / 60  # to minutes


def compute_statistics(values):
    if not values:  # Check for empty values list
        return {"mean": None, "median": None, "std": None, "min": None, "min_lang": None, "max": None, "max_lang": None}

    scores, langs = zip(*values)  # Unpack scores and languages
    min_index = np.argmin(scores)
    max_index = np.argmax(scores)
    return {
        "mean": np.mean(scores),
        "median": np.median(scores),
        "std": np.std(scores),
        "min": scores[min_index],
        "min_lang": langs[min_index],
        "max": scores[max_index],
        "max_lang": langs[max_index],
    }


def main(args):
    save_model_path = f"rcds/distilbert-SBD-{args.lang_support}-{args.type}_s{args.stride}"
    save_str = f"{save_model_path.replace('/','_')}"

    eval_data = torch.load(args.eval_data_path)

    save_str += f"{args.save_suffix}"

    # first, logits for everything.
    f, total_test_time = load_or_compute_logits(args, eval_data, save_str)

    # now, compute scores.
    results = {}
    clfs = {}
    if args.return_indices:
        indices = {}

    u_scores = []

    for lang_code, dsets in tqdm(eval_data.items()):
        if args.include_langs is not None and lang_code not in args.include_langs:
            continue

        print(f"Predicting {lang_code}...")
        results[lang_code] = {}
        clfs[lang_code] = {}
        if args.return_indices:
            indices[lang_code] = {}

        for dataset_name, dataset in dsets["sentence"].items():
            sentences = dataset["data"][: args.max_n_test_sentences]
            if not sentences:
                continue
            # check if f[lang_code][dataset_name] exists
            if lang_code not in f or dataset_name not in f[lang_code]:
                continue

            clf = [None, None, None, 0.5]

            if isinstance(sentences[0], list):
                acc_u = []
                score_u = []
                u_indices, true_indices = [], []
                length = 0
                for i, short_seq in enumerate(sentences):
                    start, end = f[lang_code][dataset_name]["test_logit_lengths"][i]
                    single_score_u, _, info, cur_u_indices, _ = evaluate_mixture(
                        lang_code,
                        f[lang_code][dataset_name]["test_logits"][:][start:end],
                        list(short_seq),
                        args.return_indices,
                        0,
                        *clf,
                    )
                    score_u.append(single_score_u)
                    acc_u.append(info["info_newline"]["correct_pairwise"])
                    # indices: accumulate from start
                    u_indices.append(
                        [idx + start for idx in cur_u_indices["pred_indices"]] if cur_u_indices["pred_indices"] else []
                    )
                    true_indices.append(
                        [idx + start for idx in cur_u_indices["true_indices"]] if cur_u_indices["true_indices"] else []
                    )
                    length += cur_u_indices["length"] - 1

            else:
                raise NotImplementedError

            if isinstance(sentences[0], list):
                score_u = np.mean(score_u)
                acc_u = np.mean(acc_u)

                results[lang_code][dataset_name] = {
                    "u": score_u,
                    "acc_u": acc_u,
                }
            else:
                raise NotImplementedError

            if args.return_indices:
                if isinstance(sentences[0], list):
                    indices[lang_code][dataset_name] = {
                        "u": u_indices,
                        "true_indices": true_indices,
                        "length": length,
                    }
                else:
                    raise NotImplementedError

            if score_u is not None:
                u_scores.append((score_u, lang_code))

            # just for printing
            print(f"{lang_code} {dataset_name} {score_u:.3f}")

    # Compute statistics for each metric across all languages
    results_avg = {
        "u": compute_statistics(u_scores),
        "include_langs": args.include_langs,
    }

    json.dump(
        results,
        open(
            Constants.CACHE_DIR / "law_bert" / f"{save_str}.json",
            "w",
        ),
        indent=4,
    )
    print(Constants.CACHE_DIR / "law_bert" / f"{save_str}.json")

    # Write results_avg to JSON
    json.dump(
        results_avg,
        open(
            Constants.CACHE_DIR / "law_bert" / f"{save_str}_AVG.json",
            "w",
        ),
        indent=4,
    )
    if args.return_indices:
        json.dump(
            indices,
            open(
                Constants.CACHE_DIR / "law_bert" / f"{save_str}_IDX.json",
                "w",
            ),
            default=int,
            # indent=4,
        )
        print(Constants.CACHE_DIR / "law_bert" / f"{save_str}_IDX.json")
        print("Indices saved to file.")

    return results, results_avg, total_test_time


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()
    results, results_avg, total_test_time = main(args)
    print(total_test_time)



import copy
import json
from dataclasses import dataclass
from typing import List, Tuple
import os
import time
import random
import sys
import logging

import h5py
import torch
from datasets import load_dataset
from tqdm.auto import tqdm
from transformers import AutoModelForTokenClassification, HfArgumentParser
import numpy as np
import adapters

import wtpsplit.models  # noqa: F401
from wtpsplit.evaluation import evaluate_mixture, get_labels, train_mixture
from wtpsplit.evaluation.intrinsic_baselines import split_language_data
from wtpsplit.extract import PyTorchWrapper
from wtpsplit.extract_batched import extract_batched
from wtpsplit.utils import Constants, token_to_char_probs
from wtpsplit.evaluation.adapt import compute_statistics

logger = logging.getLogger()
logger.setLevel(logging.INFO)


@dataclass
class Args:
    model_path: str
    adapter_path: str = None
    # eval data in the format:
    # {
    #    "<lang_code>": {
    #        "sentence": {
    #            "<dataset_name>": {
    #                 "meta": {
    #                     "train_data": ["train sentence 1", "train sentence 2"]
    #                 },
    #                 "data": ["test sentence 1", "test sentence 2"]
    #            }
    #        }
    #    }
    # }
    eval_data_path: str = "data/all_data.pth"
    valid_text_path: str = None  # "data/sentence/valid.parquet"
    device: str = "cpu"
    block_size: int = 512
    batch_size: int = 128
    include_langs: List[str] = None
    threshold: float = 0.01
    max_n_train_sentences: int = 1_000
    max_n_test_sentences: int = sys.maxsize
    save_suffix: str = ""
    skip_adaptation: bool = False
    keep_logits: bool = True
    skip_corrupted: bool = True
    skip_punct: bool = True
    return_indices: bool = True
    clf_from_scratch: bool = False

    # k_mer-specific args
    # k=2 means pairwise, k=3 triplets, ...
    k: int = 2
    max_n_samples: int = sys.maxsize
    sample_pct: float = 0.5
    min_k_mer_length: int = 0


def process_logits_k_mers(pairs, model, lang_code, block_size, batch_size, verbose=True) -> List[np.ndarray]:
    logits_list = []
    n_tokens_list = []
    # create batches of sentence pairs
    batched_k_mers = [pairs[i : i + batch_size] for i in range(0, len(pairs), batch_size)]
    for batch in tqdm(batched_k_mers, disable=not verbose):
        k_mer_texts = [Constants.SEPARATORS[lang_code].join(pair) for pair in batch]
        all_logits, offsets_mapping, tokenizer = extract_batched(
            k_mer_texts,
            model,
            lang_code=lang_code,
            block_size=block_size,
            batch_size=batch_size,
            pad_last_batch=True,
        )

        for k_mer, logit, offset_mapping in zip(k_mer_texts, all_logits, offsets_mapping):
            if "xlm" in model.config.model_type:
                tokens = tokenizer.tokenize(k_mer, verbose=False)

                # padding is also removed here (via offset_mapping)
                logits = token_to_char_probs(k_mer, tokens, logit, tokenizer, offset_mapping)
                logits_list.append(logits)
                n_tokens_list.append(len(tokens))
            else:
                if len(logit) < offset_mapping:
                    # truncated input --> pad back
                    logit = np.pad(
                        logit, ((0, offset_mapping - len(logit)), (0, 0)), "constant", constant_values=np.min(logit)
                    )
                # since we pad to equal length, we need to remove the padding
                logits_list.append(logit[:offset_mapping])

    return logits_list, n_tokens_list


def generate_k_mers(
    sentences: List[str],
    k: int,
    sample_pct: float = 1,
    max_n_samples: int = sys.maxsize,
    min_k_mer_length: int = 0,
) -> List[Tuple[str, ...]]:
    """Generate k-mers from a list of sentences.

    Args:
        sentences (List[str]): Input list of sentences.
        k (int): The number of sentences to include in each k-mer.
        sample_pct (float): Percentage of k-mers to sample.
        max_n_samples (int): Maximum number of k-mers to sample.
        min_k_mer_length (int): Minimum length of a k-mer.

    Returns:
        List[Tuple[str, ...]]: List of k-mers.
    """
    random.seed(42)
    n_k_mers = len(sentences) // k
    sample_size = min(round(n_k_mers * sample_pct), max_n_samples)

    # sample if needed
    if sample_size < n_k_mers:
        sampled_indices = set(random.sample(range(n_k_mers), sample_size))
        all_k_mers = [
            tuple(sentences[i * k + j] for j in range(k))
            for i in sampled_indices
            if sum(len(sentences[i * k + j]) for j in range(k)) > min_k_mer_length
        ]
    else:
        # all
        all_k_mers = [
            tuple(sentences[i + j] for j in range(k))
            for i in range(0, len(sentences) - k + 1, k)
            if sum(len(sentences[i + j]) for j in range(k)) > min_k_mer_length
        ]

    return all_k_mers


def load_or_compute_logits(args, model, eval_data, valid_data=None, save_str: str = None):
    logits_path = Constants.CACHE_DIR / "intrinsic_pairwise" / f"{save_str}.h5"

    if not os.path.exists(Constants.CACHE_DIR / "intrinsic_pairwise"):
        os.makedirs(Constants.CACHE_DIR / "intrinsic_pairwise")

    total_test_time = 0

    start_time = time.time()
    with h5py.File(logits_path, "a") as f, torch.no_grad():
        for lang_code in Constants.LANGINFO.index:
            if args.include_langs is not None and lang_code not in args.include_langs:
                continue
            print(f"Processing {lang_code}...")
            if lang_code not in f:
                lang_group = f.create_group(lang_code)
            else:
                lang_group = f[lang_code]

            # eval data
            for dataset_name, dataset in eval_data[lang_code]["sentence"].items():
                if args.skip_corrupted and "corrupted" in dataset_name and "ted2020" not in dataset_name:
                    continue
                if "-" in lang_code and "canine" in args.model_path and "no-adapters" not in args.model_path:
                    # code-switched data: eval 2x
                    lang_code = lang_code.split("_")[1].lower()
                try:
                    if args.adapter_path:
                        if args.clf_from_scratch:
                            model.model.classifier = torch.nn.Linear(model.model.classifier.in_features, 1)
                        elif args.model_path == "xlm-roberta-base" or args.model_path == "xlm-roberta-large":
                            # we train XLM-R using our wrapper, needs to be adapted for adapters to be loaded
                            model.model.classifier = torch.nn.Linear(
                                model.model.classifier.in_features,
                                1,
                            )
                            model.model.__class__.__name__ = "SubwordXLMForTokenClassification"
                        dataset_load_name = dataset_name
                        model.model.load_adapter(
                            args.adapter_path + "/" + dataset_load_name + "/" + lang_code,
                            set_active=True,
                            with_head=True,
                            load_as="text",
                        )
                except Exception as e:
                    print(f"Error loading adapter for {dataset_name} in {lang_code}: {e}")
                    continue
                print(dataset_name)
                if dataset_name not in lang_group:
                    dset_group = lang_group.create_group(dataset_name)
                else:
                    dset_group = lang_group[dataset_name]

                if "test_logits" not in dset_group:
                    test_sentences = dataset["data"][: args.max_n_test_sentences]
                    if not test_sentences:
                        continue
                    if isinstance(test_sentences[0], list):
                        continue
                    all_pairs_test = generate_k_mers(
                        test_sentences,
                        k=args.k,
                        sample_pct=args.sample_pct,
                        max_n_samples=args.max_n_samples,
                        min_k_mer_length=args.min_k_mer_length,
                    )

                    start_time = time.time()
                    test_logits, test_n_logits = process_logits_k_mers(
                        all_pairs_test,
                        model,
                        lang_code,
                        args.block_size,
                        args.batch_size,
                    )
                    end_time = time.time()

                    test_logit_lengths = []
                    # store start and end indices for each pair, used later to slice the logits
                    all_logit_lengths = np.append(0, np.cumsum([len(logits) for logits in test_logits]))
                    # append tuple of start and end indices for each pair
                    for i in range(len(test_logits)):
                        test_logit_lengths.append((all_logit_lengths[i], all_logit_lengths[i + 1] - 1))

                    test_logits = np.concatenate(test_logits)
                    total_test_time += end_time - start_time

                    # get_labels returns 2nd label at end of seq, which we do not want.
                    # label is at position -2 --> remove and add back 0 to end of sequence
                    test_labels = [
                        np.append(get_labels(lang_code, pair, after_space=False)[:-2], 0) for pair in all_pairs_test
                    ]

                    # flatten; append 0 eos to account for later indexing/slicing
                    test_labels = np.append(np.concatenate(test_labels), 0)
                    assert len(test_labels) == len(test_logits) + 1

                    dset_group.create_dataset("test_logits", data=test_logits)
                    dset_group.create_dataset("test_labels", data=test_labels)
                    dset_group.create_dataset("test_logit_lengths", data=test_logit_lengths)
                    dset_group.create_dataset("test_n_logits", data=test_n_logits)

                train_sentences = dataset["meta"].get("train_data")
                if train_sentences is not None and "train_logits" not in dset_group and not args.skip_adaptation:
                    train_sentences = train_sentences[: args.max_n_train_sentences]
                    all_pairs_train = generate_k_mers(
                        train_sentences,
                        k=args.k,
                        sample_pct=args.sample_pct,
                        max_n_samples=args.max_n_samples,
                        min_k_mer_length=args.min_k_mer_length,
                    )

                    train_logits, train_n_logits = process_logits_k_mers(
                        all_pairs_train, model, lang_code, args.block_size, args.batch_size
                    )
                    train_logits = np.concatenate(train_logits)

                    train_labels = [
                        np.append(get_labels(lang_code, pair, after_space=False)[:-2], 0) for pair in all_pairs_train
                    ]
                    train_labels = np.append(np.concatenate(train_labels), 0)
                    assert len(train_labels) == len(train_logits) + 1

                    dset_group.create_dataset("train_logits", data=train_logits)
                    dset_group.create_dataset("train_labels", data=train_labels)
                    dset_group.create_dataset("train_n_logits", data=train_n_logits)

    end_time = time.time()
    return h5py.File(logits_path, "r"), total_test_time / 60  # to minutes


def main(args):
    save_model_path = args.model_path
    if args.adapter_path:
        save_model_path = args.adapter_path
    save_str = f"{save_model_path.replace('/','_')}_b{args.block_size}_k{args.k}{args.save_suffix}"

    print(save_str)
    eval_data = torch.load(args.eval_data_path)
    if "canine" in args.model_path and "no-adapters" not in args.model_path:
        eval_data = split_language_data(eval_data)
    if args.valid_text_path is not None:
        valid_data = load_dataset("parquet", data_files=args.valid_text_path, split="train")
    else:
        valid_data = None

    print("Loading model...")
    model = PyTorchWrapper(AutoModelForTokenClassification.from_pretrained(args.model_path).to(args.device))
    if args.adapter_path:
        model_type = model.model.config.model_type
        # adapters need xlm-roberta as model type.
        model.model.config.model_type = "xlm-roberta"
        adapters.init(model.model)
        # reset model type (used later)
        model.model.config.model_type = model_type

    # first, logits for everything.
    f, total_test_time = load_or_compute_logits(args, model, eval_data, valid_data, save_str)
    save_str += f"_u{args.threshold}"

    # now, compute the intrinsic scores.
    results = {}
    clfs = {}
    if args.return_indices:
        indices = {}
    # lists to store scores for each metric across *all* languages
    u_scores, t_scores, punct_scores = [], [], []
    u_accs, t_accs, punct_accs = [], [], []
    thresholds_t, thresholds_adj = [], []

    for lang_code, dsets in tqdm(eval_data.items()):
        if args.include_langs is not None and lang_code not in args.include_langs:
            continue

        print(f"Predicting {lang_code}...")
        results[lang_code] = {}
        clfs[lang_code] = {}
        if args.return_indices:
            indices[lang_code] = {}

        for dataset_name, dataset in dsets["sentence"].items():
            sentences = dataset["data"][: args.max_n_test_sentences]
            if not sentences:
                continue
            if isinstance(sentences[0], list):
                continue
            sent_k_mers = generate_k_mers(
                sentences,
                k=args.k,
                sample_pct=args.sample_pct,
                max_n_samples=args.max_n_samples,
                min_k_mer_length=args.min_k_mer_length,
            )
            if lang_code not in f or dataset_name not in f[lang_code]:
                continue

            if "train_logits" in f[lang_code][dataset_name] and not args.skip_adaptation:
                feature_indices = None
                # it is sufficient to feed in 1 long sequence of tokens here since we only use logits for LR
                clf = train_mixture(
                    [lang_code],
                    f[lang_code][dataset_name]["train_logits"][:],
                    f[lang_code][dataset_name]["train_labels"][:],
                    features=feature_indices,
                    skip_punct=args.skip_punct,
                )
                # XXX: clf thresholds are still fitted on max. F1 score, not accuracy!
                # (but still without a positive label at the end)
                if clf[0] is not None:
                    print(clf)

                score_t = []
                score_punct = []
                # acc: average of correct 100% pairwise (or: k-mer) segmentation
                acc_t = []
                acc_punct = []

                # evaluate each pair
                for i, k_mer in enumerate(sent_k_mers):
                    start, end = f[lang_code][dataset_name]["test_logit_lengths"][i]
                    single_score_t, single_score_punct, info, t_indices, punct_indices = evaluate_mixture(
                        lang_code,
                        f[lang_code][dataset_name]["test_logits"][:][start:end],
                        list(k_mer),
                        args.return_indices,
                        0,
                        *clf,
                    )
                    score_t.append(single_score_t)
                    score_punct.append(single_score_punct)
                    acc_t.append(info["info_newline"]["correct_pairwise"] if info["info_newline"] else None)
                    acc_punct.append(info["info_transformed"]["correct_pairwise"] if info["info_transformed"] else None)

                clfs[lang_code][dataset_name] = clf

                clf = list(copy.deepcopy(clf))
                threshold_t = float(clf[-1])
                clf[-1] = args.threshold
            else:
                threshold_t = 0
                clf = [None, None, None, args.threshold]
                score_t = score_punct = None
                acc_t = acc_punct = None

            score_u = []
            acc_u = []
            thresholds = []
            u_indices, true_indices = [], []
            length = []
            for i, k_mer in enumerate(sent_k_mers):
                start, end = f[lang_code][dataset_name]["test_logit_lengths"][i]
                thresholds.append(args.threshold)
                single_score_u, _, info, cur_u_indices, _ = evaluate_mixture(
                    lang_code,
                    f[lang_code][dataset_name]["test_logits"][:][start:end],
                    list(k_mer),
                    args.return_indices,
                    0,
                    *clf,
                )
                score_u.append(single_score_u)
                acc_u.append(info["info_newline"]["correct_pairwise"])
                u_indices.append(cur_u_indices["pred_indices"] if cur_u_indices["pred_indices"] else [])
                true_indices.append(cur_u_indices["true_indices"] if cur_u_indices["true_indices"] else [])
                length.append(cur_u_indices["length"])

            score_u = np.mean(score_u)
            score_t = np.mean(score_t) if score_t and not args.skip_adaptation else None
            score_punct = (
                np.mean(score_punct) if score_punct and not (args.skip_punct or args.skip_adaptation) else None
            )
            acc_u = np.mean(acc_u)
            acc_t = np.mean(acc_t) if score_t else None
            acc_punct = np.mean(acc_punct) if score_punct else None
            threshold = np.mean(thresholds)

            results[lang_code][dataset_name] = {
                "u": score_u,
                "t": score_t,
                "punct": score_punct,
                "u_acc": acc_u,
                "t_acc": acc_t,
                "punct_acc": acc_punct,
                "threshold_t": threshold_t,
                "threshold_adj": threshold,
            }

            if args.return_indices:
                indices[lang_code][dataset_name] = {
                    "u": {"predicted_indices": u_indices, "true_indices": true_indices, "length": length},
                }
            # just for printing
            score_t = score_t or 0.0
            score_punct = score_punct or 0.0
            acc_t = acc_t or 0.0
            acc_punct = acc_punct or 0.0

            u_scores.append((score_u, lang_code))
            u_accs.append((acc_u, lang_code))
            t_scores.append((score_t, lang_code))
            t_accs.append((acc_t, lang_code))
            punct_scores.append((score_punct, lang_code))
            punct_accs.append((acc_punct, lang_code))
            thresholds_t.append((threshold_t, lang_code))
            thresholds_adj.append((threshold, lang_code))

            print(f"{lang_code} {dataset_name} {score_u:.3f} {score_t:.3f} {score_punct:.3f}")
            print(f"ACC: {acc_u:.3f} {acc_t:.3f} {acc_punct:.3f}")
            print(f"Threshold_t: {threshold_t:.3f} Threshold_adj: {threshold:.3f}")

    # Compute statistics for each metric across all languages
    results_avg = {
        "u": compute_statistics(u_scores),
        "t": compute_statistics(t_scores),
        "punct": compute_statistics(punct_scores),
        "u_acc": compute_statistics(u_accs),
        "t_acc": compute_statistics(t_accs),
        "punct_acc": compute_statistics(punct_accs),
        "threshold_t": compute_statistics(thresholds_t),
        "threshold_adj": compute_statistics(thresholds_adj),
        "include_langs": args.include_langs,
    }

    json.dump(
        results,
        open(
            Constants.CACHE_DIR / "intrinsic_pairwise" / f"{save_str}.json",
            "w",
        ),
        indent=4,
    )

    json.dump(
        results_avg,
        open(
            Constants.CACHE_DIR / "intrinsic_pairwise" / f"{save_str}_AVG.json",
            "w",
        ),
        indent=4,
    )

    if args.return_indices:
        json.dump(
            indices,
            open(
                Constants.CACHE_DIR / "intrinsic_pairwise" / f"{save_str}_IDX.json",
                "w",
            ),
            default=int,
            indent=4,
        )
        print(Constants.CACHE_DIR / "intrinsic_pairwise" / f"{save_str}_IDX.json")
        print("Indices saved to file.")
    if not args.keep_logits:
        os.remove(f.filename)

    print(results_avg)
    print(save_str)
    return results, results_avg, total_test_time


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()
    results, results_avg, total_test_time = main(args)
    print(total_test_time)



from transformers import AutoConfig, BertConfig, CanineConfig, XLMRobertaConfig


class LACanineConfig(CanineConfig):
    model_type = "la-canine"

    def __init__(
        self,
        n_languages=None,
        ngram_order=1,
        bottleneck_factor=2,
        language_adapter="on",
        lookahead=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.n_languages = n_languages
        self.ngram_order = ngram_order
        self.language_adapter = language_adapter  # 'on', 'off', 'shared'
        self.bottleneck_factor = bottleneck_factor

        self.lookahead = lookahead
        self.lookahead_block_size = 1


class BertCharConfig(BertConfig):
    model_type = "bert-char"

    def __init__(
        self,
        num_hash_buckets=8192,
        num_hash_functions=8,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.num_hash_buckets = num_hash_buckets
        self.num_hash_functions = num_hash_functions


class SubwordXLMConfig(XLMRobertaConfig):
    """Config for XLM-R. Used for token-level training, i.e., SaT models.

    Args:
        XLMRobertaConfig: Base class.
    """

    model_type = "xlm-token"
    mixture_name = "xlm-token"

    def __init__(
        self,
        lookahead=None,
        lookahead_split_layers=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.mixture_name = "xlm-token"
        self.lookahead = lookahead
        self.lookahead_split_layers = lookahead_split_layers


AutoConfig.register("bert-char", BertCharConfig)
AutoConfig.register("la-canine", LACanineConfig)
AutoConfig.register("xlm-token", SubwordXLMConfig)



import os
from typing import Dict

import numpy as np
import torch
import transformers
from torch import nn
from torch.optim.lr_scheduler import LambdaLR
from transformers import PreTrainedModel
from transformers.trainer import (
    ALL_LAYERNORM_LAYERS,
    TRAINING_ARGS_NAME,
    WEIGHTS_NAME,
    DataLoader,
    EvalLoopOutput,
    IterableDatasetShard,
    List,
    Optional,
    ShardedDDPOption,
    deepspeed_init,
    denumpify_detensorize,
    find_batch_size,
    get_parameter_names,
    has_length,
    is_sagemaker_mp_enabled,
    is_torch_tpu_available,
    logger,
    nested_concat,
    nested_numpify,
    nested_truncate,
)
from transformers.modeling_utils import unwrap_model

from wtpsplit.train.utils import Model

if is_torch_tpu_available(check_device=False):
    import torch_xla.core.xla_model as xm  # noqa: F401
    import torch_xla.debug.metrics as met  # noqa: F401
    import torch_xla.distributed.parallel_loader as pl  # noqa: F401


class Trainer(transformers.Trainer):
    def get_param_groups(self):
        opt_model = self.model_wrapped if is_sagemaker_mp_enabled() else self.model

        decay_parameters = get_parameter_names(opt_model, ALL_LAYERNORM_LAYERS)
        decay_parameters = [name for name in decay_parameters if "bias" not in name]

        lang_adapter_parameters = [
            name
            for name, _ in opt_model.named_parameters()
            if "lang" in name or "classifier" in name or "projection" in name or "HashBucketCodepointEmbedder" in name
        ]

        optimizer_grouped_parameters = [
            {
                "params": [
                    p
                    for n, p in opt_model.named_parameters()
                    if n in decay_parameters and n not in lang_adapter_parameters
                ],
                "weight_decay": self.args.weight_decay,
            },
            {
                "params": [
                    p
                    for n, p in opt_model.named_parameters()
                    if n not in decay_parameters and n not in lang_adapter_parameters
                ],
                "weight_decay": 0.0,
            },
        ]

        if len(lang_adapter_parameters) > 0:
            optimizer_grouped_parameters.extend(
                [
                    {
                        "params": [
                            p
                            for n, p in opt_model.named_parameters()
                            if n in decay_parameters and n in lang_adapter_parameters
                        ],
                        "weight_decay": self.args.weight_decay,
                    },
                    {
                        "params": [
                            p
                            for n, p in opt_model.named_parameters()
                            if n not in decay_parameters and n in lang_adapter_parameters
                        ],
                        "weight_decay": 0.0,
                    },
                ]
            )

        return optimizer_grouped_parameters

    def create_optimizer(self):
        """
        Setup the optimizer.
        We provide a reasonable default that works well. If you want to use something else, you can pass a tuple in the
        Trainer's init through `optimizers`, or subclass and override this method in a subclass.
        """
        opt_model = self.model_wrapped if is_sagemaker_mp_enabled() else self.model

        if self.optimizer is None:
            optimizer_grouped_parameters = self.get_param_groups()

            optimizer_cls, optimizer_kwargs = Trainer.get_optimizer_cls_and_kwargs(self.args)

            if self.sharded_ddp == ShardedDDPOption.SIMPLE:
                raise NotImplementedError()
            else:
                self.optimizer = optimizer_cls(optimizer_grouped_parameters, **optimizer_kwargs)
                if optimizer_cls.__name__ == "Adam8bit":
                    import bitsandbytes

                    manager = bitsandbytes.optim.GlobalOptimManager.get_instance()

                    for module in opt_model.modules():
                        if isinstance(module, nn.Embedding):
                            manager.register_module_override(module, "weight", {"optim_bits": 32})
                            logger.debug(f"bitsandbytes: will optimize {module} in fp32")

        if is_sagemaker_mp_enabled():
            raise NotImplementedError()

        return self.optimizer

    def create_scheduler(self, num_training_steps: int, optimizer: torch.optim.Optimizer = None):
        if self.lr_scheduler is None:
            warmup_steps = self.args.get_warmup_steps(self.args.max_steps)

            # MODIFIED: add lang adapter lr scheduler (wtp only)
            def lr_lambda(current_step: int):
                return 0.0
                if current_step < self.args.adapter_warmup_steps:
                    return 0.0

                step = current_step - self.args.adapter_warmup_steps

                if step < warmup_steps:
                    return float(step) / float(warmup_steps)
                else:
                    return max(
                        0.0,
                        float(num_training_steps - step) / float(max(1, num_training_steps - warmup_steps)),
                    )

            def lr_lambda_lang_adapter(current_step: int):
                return 0.0
                if current_step < self.args.adapter_warmup_steps:
                    return self.args.adapter_lr_multiplier
                else:
                    return self.args.adapter_lr_multiplier * lr_lambda(current_step)

            self.lr_scheduler = LambdaLR(
                optimizer=self.optimizer if optimizer is None else optimizer,
                lr_lambda=[
                    lr_lambda,
                    lr_lambda,
                    lr_lambda_lang_adapter,
                    lr_lambda_lang_adapter,
                ]
                if len(self.get_param_groups()) == 4
                else [lr_lambda, lr_lambda],
            )

        return self.lr_scheduler

    def _maybe_log_save_evaluate(self, tr_loss, model, trial, epoch, ignore_keys_for_eval):
        if self.control.should_log:
            if is_torch_tpu_available():
                xm.mark_step()

            logs: Dict[str, float] = {}

            # all_gather + mean() to get average loss over all processes
            tr_loss_scalar = self._nested_gather(tr_loss).mean().item()

            # reset tr_loss to zero
            tr_loss -= tr_loss

            logs["loss"] = round(
                tr_loss_scalar / (self.state.global_step - self._globalstep_last_logged),
                4,
            )
            (
                logs["learning_rate"],
                _,
                logs["learning_rate_adapter"],
                _,
            ) = self.lr_scheduler.get_last_lr()

            self._total_loss_scalar += tr_loss_scalar
            self._globalstep_last_logged = self.state.global_step
            self.store_flos()

            self.log(logs)

        metrics = None
        if self.control.should_evaluate:
            metrics = self.evaluate(ignore_keys=ignore_keys_for_eval)
            self._report_to_hp_search(trial, self.state.global_step, metrics)

        if self.control.should_save:
            self._save_checkpoint(model, trial, metrics=metrics)
            self.control = self.callback_handler.on_save(self.args, self.state, self.control)

    def evaluation_loop(
        self,
        dataloader: DataLoader,
        description: str,
        prediction_loss_only: Optional[bool] = None,
        ignore_keys: Optional[List[str]] = None,
        metric_key_prefix: str = "eval",
    ) -> EvalLoopOutput:
        """
        Prediction/evaluation loop, shared by `Trainer.evaluate()` and `Trainer.predict()`.

        Works both with or without labels.
        """
        args = self.args

        prediction_loss_only = prediction_loss_only if prediction_loss_only is not None else args.prediction_loss_only

        # if eval is called w/o train init deepspeed here
        if args.deepspeed and not self.deepspeed:
            # XXX: eval doesn't have `resume_from_checkpoint` arg but we should be able to do eval
            # from the checkpoint eventually
            deepspeed_engine, _, _ = deepspeed_init(
                self, num_training_steps=0, resume_from_checkpoint=None, inference=True
            )
            self.model = deepspeed_engine.module
            self.model_wrapped = deepspeed_engine
            self.deepspeed = deepspeed_engine

        model = self._wrap_model(self.model, training=False, dataloader=dataloader)

        # if full fp16 or bf16 eval is wanted and this ``evaluation`` or ``predict`` isn't called
        # while ``train`` is running, cast it to the right dtype first and then put on device
        if not self.is_in_train:
            if args.fp16_full_eval:
                model = model.to(dtype=torch.float16, device=args.device)
            elif args.bf16_full_eval:
                model = model.to(dtype=torch.bfloat16, device=args.device)

        batch_size = self.args.eval_batch_size

        logger.warning(f"***** Running {description} *****")
        if has_length(dataloader):
            logger.warning(f"  Num examples = {self.num_examples(dataloader)}")
        else:
            logger.warning("  Num examples: Unknown")
        logger.warning(f"  Batch size = {batch_size}")

        model.eval()

        self.callback_handler.eval_dataloader = dataloader
        # Do this before wrapping.
        eval_dataset = getattr(dataloader, "dataset", None)

        if is_torch_tpu_available():
            dataloader = pl.ParallelLoader(dataloader, [args.device]).per_device_loader(args.device)

        if args.past_index >= 0:
            self._past = None

        # Initialize containers
        # losses/preds/labels on GPU/TPU (accumulated for eval_accumulation_steps)
        losses_host = None
        preds_host = None
        labels_host = None
        inputs_host = None

        # losses/preds/labels on CPU (final containers)
        all_losses = None
        all_preds = None
        all_labels = None
        all_inputs = None
        # Will be useful when we have an iterable dataset so don't know its length.

        observed_num_examples = 0
        # Main evaluation loop
        for step, inputs in enumerate(dataloader):
            # Update the observed num examples
            observed_batch_size = find_batch_size(inputs)
            if observed_batch_size is not None:
                observed_num_examples += observed_batch_size
                # For batch samplers, batch_size is not known by the dataloader in advance.
                if batch_size is None:
                    batch_size = observed_batch_size

            # Prediction step
            loss, logits, labels = self.prediction_step(model, inputs, prediction_loss_only, ignore_keys=ignore_keys)
            inputs_decode = self._prepare_input(inputs["input_ids"]) if args.include_inputs_for_metrics else None

            if is_torch_tpu_available():
                xm.mark_step()

            # Update containers on host
            if loss is not None:
                losses = self._nested_gather(loss.repeat(batch_size))
                losses_host = losses if losses_host is None else torch.cat((losses_host, losses), dim=0)
            if labels is not None:
                labels = self._pad_across_processes(labels)
                labels = self._nested_gather(labels)
                labels_host = labels if labels_host is None else nested_concat(labels_host, labels, padding_index=-100)
            if inputs_decode is not None:
                inputs_decode = self._pad_across_processes(inputs_decode)
                inputs_decode = self._nested_gather(inputs_decode)
                inputs_host = (
                    inputs_decode
                    if inputs_host is None
                    else nested_concat(inputs_host, inputs_decode, padding_index=-100)
                )
            if logits is not None:
                logits = self._pad_across_processes(logits)
                logits = self._nested_gather(logits)
                if self.preprocess_logits_for_metrics is not None:
                    logits = self.preprocess_logits_for_metrics(logits, labels)
                preds_host = logits if preds_host is None else nested_concat(preds_host, logits, padding_index=-100)
            self.control = self.callback_handler.on_prediction_step(args, self.state, self.control)

            # Gather all tensors and put them back on the CPU if we have done enough accumulation steps.
            if args.eval_accumulation_steps is not None and (step + 1) % args.eval_accumulation_steps == 0:
                if losses_host is not None:
                    losses = nested_numpify(losses_host)
                    all_losses = losses if all_losses is None else np.concatenate((all_losses, losses), axis=0)
                if preds_host is not None:
                    logits = nested_numpify(preds_host)
                    all_preds = logits if all_preds is None else nested_concat(all_preds, logits, padding_index=-100)
                if inputs_host is not None:
                    inputs_decode = nested_numpify(inputs_host)
                    all_inputs = (
                        inputs_decode
                        if all_inputs is None
                        else nested_concat(all_inputs, inputs_decode, padding_index=-100)
                    )
                if labels_host is not None:
                    labels = nested_numpify(labels_host)
                    all_labels = labels if all_labels is None else nested_concat(all_labels, labels, padding_index=-100)

                # Set back to None to begin a new accumulation
                losses_host, preds_host, inputs_host, labels_host = (
                    None,
                    None,
                    None,
                    None,
                )

        if args.past_index and hasattr(self, "_past"):
            # Clean the state at the end of the evaluation loop
            delattr(self, "_past")

        # Gather all remaining tensors and put them back on the CPU
        if losses_host is not None:
            losses = nested_numpify(losses_host)
            all_losses = losses if all_losses is None else np.concatenate((all_losses, losses), axis=0)
        if preds_host is not None:
            logits = nested_numpify(preds_host)
            all_preds = logits if all_preds is None else nested_concat(all_preds, logits, padding_index=-100)
        if inputs_host is not None:
            inputs_decode = nested_numpify(inputs_host)
            all_inputs = (
                inputs_decode if all_inputs is None else nested_concat(all_inputs, inputs_decode, padding_index=-100)
            )
        if labels_host is not None:
            labels = nested_numpify(labels_host)
            all_labels = labels if all_labels is None else nested_concat(all_labels, labels, padding_index=-100)

        # Number of samples
        if has_length(eval_dataset):
            num_samples = len(eval_dataset)
        # The instance check is weird and does not actually check for the type, but whether the dataset has the right
        # methods. Therefore we need to make sure it also has the attribute.
        elif isinstance(eval_dataset, IterableDatasetShard) and hasattr(eval_dataset, "num_examples"):
            num_samples = eval_dataset.num_examples
        else:
            if has_length(dataloader):
                num_samples = self.num_examples(dataloader)
            else:  # both len(dataloader.dataset) and len(dataloader) fail
                num_samples = observed_num_examples

        # Number of losses has been rounded to a multiple of batch_size and in a distributed training, the number of
        # samplers has been rounded to a multiple of batch_size, so we truncate.
        if all_losses is not None:
            all_losses = all_losses[:num_samples]
        if all_preds is not None:
            all_preds = nested_truncate(all_preds, num_samples)
        if all_labels is not None:
            all_labels = nested_truncate(all_labels, num_samples)
        if all_inputs is not None:
            all_inputs = nested_truncate(all_inputs, num_samples)

        # Metrics!
        # MODIFIED: always compute metrics
        if self.compute_metrics is not None:
            metrics = self.compute_metrics(self)
        else:
            metrics = {}

        # To be JSON-serializable, we need to remove numpy types or zero-d tensors
        metrics = denumpify_detensorize(metrics)

        if all_losses is not None:
            metrics[f"{metric_key_prefix}_loss"] = all_losses.mean().item()

        # Prefix all keys with metric_key_prefix + '_'
        for key in list(metrics.keys()):
            if not key.startswith(f"{metric_key_prefix}_"):
                metrics[f"{metric_key_prefix}_{key}"] = metrics.pop(key)

        return EvalLoopOutput(
            predictions=all_preds,
            label_ids=all_labels,
            metrics=metrics,
            num_samples=num_samples,
        )

    def _save_tpu(self, output_dir: Optional[str] = None):
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        logger.warning(f"Saving model checkpoint to {output_dir}")

        if xm.is_master_ordinal():
            os.makedirs(output_dir, exist_ok=True)
            torch.save(self.args, os.path.join(output_dir, TRAINING_ARGS_NAME))

        # Save a trained model and configuration using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        xm.rendezvous("saving_checkpoint")
        if isinstance(self.model, Model):
            actual_model = self.model.backbone
        else:
            actual_model = self.model
        if not isinstance(actual_model, PreTrainedModel):
            if isinstance(unwrap_model(actual_model), PreTrainedModel):
                unwrap_model(actual_model).save_pretrained(
                    output_dir,
                    is_main_process=self.args.should_save,
                    state_dict=actual_model.state_dict(),
                    save_function=xm.save,
                )
            else:
                logger.warning("Trainer.model is not a `PreTrainedModel`, only saving its state dict.")
                state_dict = actual_model.state_dict()
                xm.save(state_dict, os.path.join(output_dir, WEIGHTS_NAME))
        else:
            actual_model.save_pretrained(output_dir, is_main_process=self.args.should_save, save_function=xm.save)
        if self.tokenizer is not None and self.args.should_save:
            self.tokenizer.save_pretrained(output_dir)



import math
from dataclasses import dataclass
from functools import partial

import numpy as np
import torch
import wandb
from datasets import load_dataset
from sklearn.metrics import f1_score
from torch import nn
from tqdm.auto import tqdm
from transformers import (
    AutoModel,
    AutoModelForTokenClassification,
    AutoTokenizer,
    CanineTokenizer,
    HfArgumentParser,
    Trainer,
    TrainingArguments,
)

from wtpsplit.models import LACanineConfig, LACanineForTokenClassification, LACanineModel
from wtpsplit.utils import Constants


@dataclass
class Args:
    name: str
    lang: str
    model_path: str = None
    baseline: str = None
    block_size: int = 256
    stride: int = 32
    use_lang_adapter: bool = False
    use_deep_punctuation: bool = False


class DeepPunctuation(nn.Module):
    def __init__(self, backbone, freeze_bert=False, lstm_dim=-1):
        super(DeepPunctuation, self).__init__()
        self.output_dim = 4
        self.bert_layer = backbone
        # Freeze bert layers
        if freeze_bert:
            for p in self.bert_layer.parameters():
                p.requires_grad = False
        bert_dim = backbone.config.hidden_size
        if lstm_dim == -1:
            hidden_size = bert_dim
        else:
            hidden_size = lstm_dim
        self.lstm = nn.LSTM(
            input_size=bert_dim,
            hidden_size=hidden_size,
            num_layers=1,
            bidirectional=True,
        )
        self.linear = nn.Linear(in_features=hidden_size * 2, out_features=self.output_dim)

    def forward(self, input_ids, attention_mask=None, labels=None):
        # (B, N, E) -> (B, N, E)
        x = self.bert_layer(input_ids, attention_mask=attention_mask)[0]
        # (B, N, E) -> (N, B, E)
        x = torch.transpose(x, 0, 1)
        x, (_, _) = self.lstm(x)
        # (N, B, E) -> (B, N, E)
        x = torch.transpose(x, 0, 1)
        x = self.linear(x)

        if labels is not None:
            loss = nn.CrossEntropyLoss()(x.view(-1, 4), labels.view(-1))

        return {"logits": x, "loss": loss}


def to_strided(input_ids, labels, offset_mapping, tokenizer, stride, block_size, add_sep_cls):
    if add_sep_cls:
        block_size -= 2
        stride -= 2

    input_id_windows = []
    label_windows = []
    offset_mapping_windows = []

    start_positions = [i * stride for i in range(len(input_ids) // stride - block_size // stride)]
    if start_positions[-1] + block_size != len(input_ids):
        start_positions[-1] = len(input_ids) - block_size

    for start in start_positions:
        if add_sep_cls:
            input_id_windows.append(
                [tokenizer.cls_token_id] + input_ids[start : start + block_size] + [tokenizer.sep_token_id]
            )
            label_windows.append([0] + labels[start : start + block_size] + [0])
            offset_mapping_windows.append([(0, 0)] + offset_mapping[start : start + block_size] + [(0, 0)])
        else:
            input_id_windows.append(input_ids[start : start + block_size])
            label_windows.append(labels[start : start + block_size])
            offset_mapping_windows.append(offset_mapping[start : start + block_size])

    return input_id_windows, label_windows, offset_mapping_windows


def process(data, tokenizer, stride, block_size, add_sep_cls, fix_space, return_chars=False):
    char_labels = []

    label_dict = {
        "O": 0,
        "QUESTION": 1,
        "PERIOD": 2,
        "COMMA": 3,
    }

    text = ""

    for i, row in enumerate(data["text"]):
        token, label = row.split("\t")

        if not fix_space or (i > 0 and token not in {"'m", "n't", "'ll", "'s"}):
            text += " "
            char_labels.append(0)

        text += token
        char_labels.extend([0] * len(token))
        char_labels[-1] = label_dict[label]

    if isinstance(tokenizer, CanineTokenizer):
        encoding = tokenizer(text, add_special_tokens=False)

        assert encoding["input_ids"] == [ord(c) for c in text]
        offset_mapping = [(i, i + 1) for i in range(len(text))]
    else:
        encoding = tokenizer(text, return_offsets_mapping=True, add_special_tokens=False)
        offset_mapping = encoding["offset_mapping"]

    labels = [char_labels[x[1] - 1] for x in offset_mapping]

    if return_chars:
        return {"labels": char_labels}

    input_id_windows, label_windows, offset_mapping_windows = to_strided(
        encoding["input_ids"],
        labels,
        offset_mapping,
        tokenizer,
        stride,
        block_size,
        add_sep_cls,
    )

    return {
        "input_ids": input_id_windows,
        "labels": label_windows,
        "offset_mapping": offset_mapping_windows,
    }


def load_iwslt(path, tokenizer, args, fix_space=True):
    dataset = load_dataset("text", data_files=str(path), split="train")
    dataset = dataset.map(
        partial(
            process,
            tokenizer=tokenizer,
            stride=args.stride,
            block_size=args.block_size,
            add_sep_cls=args.add_sep_cls,
            fix_space=fix_space,
        ),
        batched=True,
        batch_size=len(dataset),
        remove_columns=["text"],
    )

    char_dataset = load_dataset("text", data_files=str(path), split="train")
    char_labels = char_dataset.map(
        partial(
            process,
            tokenizer=tokenizer,
            stride=args.stride,
            block_size=args.block_size,
            add_sep_cls=args.add_sep_cls,
            fix_space=fix_space,
            return_chars=True,
        ),
        batched=True,
        batch_size=len(char_dataset),
        remove_columns=["text"],
    )["labels"]

    if args.use_lang_adapter:
        dataset = dataset.add_column("language_ids", [Constants.LANG_CODE_TO_INDEX[args.lang]] * len(dataset))

    return dataset, np.array(char_labels)


def compute_metrics(_, test_dataset, test_char_labels):
    global trainer

    model = trainer._wrap_model(trainer.model, training=False)
    batch_size = trainer.args.per_device_eval_batch_size

    if trainer.args.process_index == 0:
        char_preds = np.zeros((len(test_char_labels), 4))
        char_counts = np.zeros((len(test_char_labels)))

        n_batches = math.ceil(len(test_dataset) / batch_size)
        for i in tqdm(range(n_batches)):
            batch = test_dataset[(i * batch_size) : (i + 1) * batch_size]
            while len(batch["input_ids"]) < batch_size:
                batch["input_ids"].append([0] * len(batch["input_ids"][-1]))
                batch["labels"].append([-100] * len(batch["input_ids"][-1]))
                batch["offset_mapping"].append([(0, 0)] * len(batch["input_ids"][-1]))

            with torch.no_grad():
                output = model(
                    input_ids=torch.tensor(batch["input_ids"], device=model.device),
                    labels=torch.tensor(batch["labels"], device=model.device),
                )

                batch_logits = output["logits"].cpu().numpy()
                for logits, offsets in zip(batch_logits, batch["offset_mapping"]):
                    for l, (start, end) in zip(logits, offsets):
                        if start != end:
                            char_preds[end - 1] += l
                            char_counts[end - 1] += 1

        mask = test_char_labels != -100

        comma, period, question = (
            f1_score(char_preds[mask].argmax(-1) == 3, test_char_labels[mask] == 3),
            f1_score(char_preds[mask].argmax(-1) == 2, test_char_labels[mask] == 2),
            f1_score(char_preds[mask].argmax(-1) == 1, test_char_labels[mask] == 1),
        )

        return {
            "comma_f1": comma,
            "period_f1": period,
            "question_f1": question,
            "overall_f1": np.mean([comma, period, question]),
        }
    else:
        return {}


def _mp_fn(index):
    # For xla_spawn (TPUs)
    main()


def main():
    (args, training_args) = HfArgumentParser([Args, TrainingArguments]).parse_args_into_dataclasses()

    model_class = AutoModel if args.use_deep_punctuation else AutoModelForTokenClassification
    our_model_class = LACanineModel if args.use_deep_punctuation else LACanineForTokenClassification

    if args.baseline:
        assert not args.use_lang_adapter

        model = model_class.from_pretrained(args.baseline, num_labels=4)
        tokenizer = AutoTokenizer.from_pretrained(args.baseline)
        args.add_sep_cls = True
    else:
        config = LACanineConfig.from_pretrained(
            args.model_path,
            num_labels=4,
            language_adapter="on" if args.use_lang_adapter else "off",
        )
        model = our_model_class.from_pretrained(args.model_path, config=config, ignore_mismatched_sizes=True)
        if args.use_lang_adapter:
            model.set_language_adapters(args.lang)

        tokenizer = CanineTokenizer.from_pretrained("google/canine-s")
        args.add_sep_cls = False

    if args.use_deep_punctuation:
        model = DeepPunctuation(model)

    if args.lang == "en":
        train_dataset, _ = load_iwslt(
            Constants.ROOT_DIR / "data" / "external" / "punctuation_annotation" / "en" / "train2012",
            tokenizer=tokenizer,
            args=args,
        )
        test_dataset, test_char_labels = load_iwslt(
            Constants.ROOT_DIR / "data" / "external" / "punctuation_annotation" / "en" / "test2011",
            tokenizer=tokenizer,
            args=args,
        )
    else:
        train_dataset, _ = load_iwslt(
            Constants.ROOT_DIR / "data" / "external" / "punctuation_annotation" / "bn" / "train",
            tokenizer=tokenizer,
            args=args,
        )
        test_dataset, test_char_labels = load_iwslt(
            Constants.ROOT_DIR / "data" / "external" / "punctuation_annotation" / "bn" / "test_ref",
            tokenizer=tokenizer,
            args=args,
        )

    if "wandb" in training_args.report_to and training_args.process_index == 0:
        wandb.init(project="punct_annotation", name=args.name)
        wandb.config.update(args)
        wandb.config.update(training_args)

        # model.config.wandb_run_id = wandb.run.id

    trainer = Trainer(
        model,
        training_args,
        train_dataset=train_dataset,
        eval_dataset=test_dataset,
        compute_metrics=partial(
            compute_metrics,
            test_dataset=test_dataset,
            test_char_labels=test_char_labels,
        ),
    )

    globals().update({"trainer": trainer})

    trainer.train()


if __name__ == "__main__":
    main()



import os
from typing import Dict

import numpy as np
import torch
from torch import nn

from transformers import PreTrainedModel
from transformers.modeling_utils import unwrap_model
from transformers.trainer import (
    ALL_LAYERNORM_LAYERS,
    WEIGHTS_NAME,
    DataLoader,
    EvalLoopOutput,
    IterableDatasetShard,
    List,
    Optional,
    deepspeed_init,
    denumpify_detensorize,
    find_batch_size,
    get_parameter_names,
    has_length,
    is_sagemaker_mp_enabled,
    is_torch_tpu_available,
    nested_concat,
    nested_numpify,
    nested_truncate,
)

from wtpsplit.train.utils import Model

if is_torch_tpu_available(check_device=False):
    import torch_xla.core.xla_model as xm  # noqa: F401
    import torch_xla.debug.metrics as met  # noqa: F401
    import torch_xla.distributed.parallel_loader as pl  # noqa: F401

import re
from typing import Callable, Tuple, Union

from packaging import version
from torch.utils.data import Dataset
from transformers import Trainer, __version__
from transformers.configuration_utils import PretrainedConfig
from transformers.data.data_collator import DataCollator
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from transformers.trainer_callback import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
)
from transformers.trainer_utils import (
    EvalPrediction,
)
from transformers.training_args import TrainingArguments
from transformers.utils import (
    CONFIG_NAME,
    is_accelerate_available,
    is_apex_available,
    is_torch_tpu_available,
    logging,
)

from adapters.composition import AdapterCompositionBlock, Fuse

if is_apex_available():
    pass

if is_sagemaker_mp_enabled():
    import smdistributed.modelparallel.torch as smp

skip_first_batches = None
if is_accelerate_available():
    from accelerate import __version__ as accelerate_version

    if version.parse(accelerate_version) >= version.parse("0.16"):
        pass
logger = logging.get_logger(__name__)

TRAINING_ARGS_NAME = "training_args.bin"


class AdapterTrainer(Trainer):
    def __init__(
        self,
        model: Union[PreTrainedModel, nn.Module] = None,
        args: TrainingArguments = None,
        data_collator: Optional[DataCollator] = None,
        train_dataset: Optional[Dataset] = None,
        eval_dataset: Optional[Dataset] = None,
        tokenizer: Optional[PreTrainedTokenizerBase] = None,
        model_init: Callable[[], PreTrainedModel] = None,
        compute_metrics: Optional[Callable[[EvalPrediction], Dict]] = None,
        callbacks: Optional[List[TrainerCallback]] = None,
        adapter_names: Optional[List[List[str]]] = None,
        optimizers: Tuple[torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR] = (None, None),
        preprocess_logits_for_metrics: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
        logging_prefix="",
        skip_eval_loss: bool = False,
    ):
        super().__init__(
            model,
            args,
            data_collator,
            train_dataset,
            eval_dataset,
            tokenizer=tokenizer,
            model_init=model_init,
            compute_metrics=compute_metrics,
            callbacks=[AdapterTrainerCallback(self)] + callbacks if callbacks else [AdapterTrainerCallback(self)],
            optimizers=optimizers,
            preprocess_logits_for_metrics=preprocess_logits_for_metrics,
        )

        self.logging_prefix = logging_prefix
        self.skip_eval_loss = skip_eval_loss

        if adapter_names is not None:
            self.model.backbone.set_active_adapters(adapter_names)
        # Set the defaults for loading/ saving model & adapters
        if isinstance(self.model.backbone, PreTrainedModel):
            model_frozen = getattr(self.model.backbone.base_model, "model_frozen", False)
        else:
            model_frozen = False
        if model_frozen and self.model.backbone.active_adapters:
            # Check if training AdapterFusion
            self.train_adapter_fusion = (
                isinstance(self.model.backbone.active_adapters, Fuse)
                or isinstance(self.model.backbone.active_adapters, AdapterCompositionBlock)
                and any([isinstance(child, Fuse) for child in self.model.backbone.active_adapters.children])
            )
        if self.model.backbone.active_adapters is None:
            raise ValueError(
                "Expected a model with an active adapter setup."
                "If you want to fully finetune the model use the Trainer class."
            )
        if (self.label_names is None or len(self.label_names) < 1) and self.model.active_head is not None:
            all_label_names = set()
            for head in self.model.backbone._active_heads:
                all_label_names |= set(self.model.backbone.heads[head].get_label_names())
            self.label_names = list(all_label_names)

    def create_optimizer(self):
        """
        Setup the optimizer.

        We provide a reasonable default that works well. If you want to use something else, you can pass a tuple in the
        Trainer's init through `optimizers`, or subclass and override this method in a subclass.
        """
        opt_model = self.model_wrapped if is_sagemaker_mp_enabled() else self.model

        if self.optimizer is None:
            decay_parameters = get_parameter_names(opt_model, ALL_LAYERNORM_LAYERS)
            decay_parameters = [name for name in decay_parameters if "bias" not in name]
            if hasattr(self.model, "config") and hasattr(self.model.config, "adapters"):
                match_str = r"adapter_fusion_layer\..*\.value"
                decay_parameters = [name for name in decay_parameters if not re.match(match_str, name)]
            optimizer_grouped_parameters = [
                {
                    "params": [
                        p for n, p in opt_model.named_parameters() if (n in decay_parameters and p.requires_grad)
                    ],
                    "weight_decay": self.args.weight_decay,
                },
                {
                    "params": [
                        p for n, p in opt_model.named_parameters() if (n not in decay_parameters and p.requires_grad)
                    ],
                    "weight_decay": 0.0,
                },
            ]

            optimizer_cls, optimizer_kwargs = Trainer.get_optimizer_cls_and_kwargs(self.args)
            self.optimizer = optimizer_cls(optimizer_grouped_parameters, **optimizer_kwargs)

        if is_sagemaker_mp_enabled():
            self.optimizer = smp.DistributedOptimizer(self.optimizer)

        return self.optimizer

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        # If we are executing this function, we are the process zero, so we don't check for that.
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info(f"Saving model checkpoint to {output_dir}")
        # Save a trained model and configuration using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        if not isinstance(self.model, PreTrainedModel):
            if isinstance(unwrap_model(self.model), PreTrainedModel):
                if state_dict is None:
                    state_dict = self.model.state_dict()
                unwrap_model(self.model).save_pretrained(output_dir, state_dict=state_dict)
            else:
                logger.info("Trainer.model is not a `PreTrainedModel`, only saving its state dict.")
                if state_dict is None:
                    state_dict = self.model.state_dict()
                torch.save(state_dict, os.path.join(output_dir, WEIGHTS_NAME))
        else:
            self.model.save_all_adapters(output_dir)
            if self.train_adapter_fusion:
                self.model.save_all_adapter_fusions(output_dir)
            if hasattr(self.model, "heads"):
                self.model.save_all_heads(output_dir)
        if self.tokenizer is not None:
            self.tokenizer.save_pretrained(output_dir)

        # Good practice: save your training arguments together with the trained model
        torch.save(self.args, os.path.join(output_dir, "training_args.bin"))

    def _load_from_checkpoint(self, resume_from_checkpoint):
        args = self.args
        if os.path.isfile(os.path.join(resume_from_checkpoint, WEIGHTS_NAME)):
            logger.info(f"Loading model from {resume_from_checkpoint}).")

        if os.path.isfile(os.path.join(resume_from_checkpoint, CONFIG_NAME)):
            config = PretrainedConfig.from_json_file(os.path.join(resume_from_checkpoint, CONFIG_NAME))
            checkpoint_version = config.transformers_version
            if checkpoint_version is not None and checkpoint_version != __version__:
                logger.warn(
                    f"You are resuming training from a checkpoint trained with {checkpoint_version} of "
                    f"Transformers but your current version is {__version__}. This is not recommended and could "
                    "yield to errors or unwanted behaviors."
                )

        if args.deepspeed:
            # will be resumed in deepspeed_init
            pass
        else:
            adapter_loaded = False
            if os.path.isdir(resume_from_checkpoint):
                adapter_loaded = self._load_adapters(resume_from_checkpoint)
                self._load_adapter_fusions(resume_from_checkpoint)
                # Save all heads for a model with heads
                if hasattr(self.model, "heads"):
                    self._load_heads(resume_from_checkpoint)

            if not adapter_loaded:
                raise Exception("Can't find a valid checkpoint at {}".format(resume_from_checkpoint))

    def _load_adapters(self, resume_from_checkpoint):
        adapter_loaded = False
        for file_name in os.listdir(resume_from_checkpoint):
            if os.path.isdir(os.path.join(resume_from_checkpoint, file_name)):
                if "," not in file_name and "adapter_config.json" in os.listdir(
                    os.path.join(resume_from_checkpoint, file_name)
                ):
                    self.model.load_adapter(os.path.join(os.path.join(resume_from_checkpoint, file_name)))
                    adapter_loaded = True
        return adapter_loaded

    def _load_adapter_fusions(self, resume_from_checkpoint):
        for file_name in os.listdir(resume_from_checkpoint):
            if os.path.isdir(os.path.join(resume_from_checkpoint, file_name)):
                if "," in file_name:
                    self.model.load_adapter_fusion(os.path.join(resume_from_checkpoint, file_name))

    def _load_heads(self, resume_from_checkpoint):
        for file_name in os.listdir(resume_from_checkpoint):
            if os.path.isdir(os.path.join(resume_from_checkpoint, file_name)):
                if "," not in file_name and "head_config.json" in os.listdir(
                    os.path.join(resume_from_checkpoint, file_name)
                ):
                    self.model.load_head(os.path.join(resume_from_checkpoint, file_name))

    def _load_best_model(self):
        model = self.model_wrapped if is_sagemaker_mp_enabled() else self.model
        logger.info(
            f"Loading best adapter(s) from {self.state.best_model_checkpoint} (score: {self.state.best_metric})."
        )
        # attempt to re-load all adapters from checkpoint
        for adapter in model.adapters_config.adapters:
            adapter_dir = os.path.join(self.state.best_model_checkpoint, adapter)
            if os.path.exists(adapter_dir):
                model.load_adapter(adapter_dir)
        if self.train_adapter_fusion:
            logger.info(
                f"Loading best adapter fusion(s) from {self.state.best_model_checkpoint} (score:"
                f" {self.state.best_metric})."
            )
            # attempt to re-load all adapter fusions from checkpoint
            for fusion in model.adapters_config.fusions:
                fusion_dir = os.path.join(self.state.best_model_checkpoint, fusion)
                if os.path.exists(fusion_dir):
                    model.load_adapter_fusion(fusion_dir)
        model.to(self.args.device)

    def evaluation_loop(
        self,
        dataloader: DataLoader,
        description: str,
        prediction_loss_only: Optional[bool] = None,
        ignore_keys: Optional[List[str]] = None,
        metric_key_prefix: str = "eval",
    ) -> EvalLoopOutput:
        """
        Prediction/evaluation loop, shared by `Trainer.evaluate()` and `Trainer.predict()`.

        Works both with or without labels.
        """
        args = self.args

        prediction_loss_only = prediction_loss_only if prediction_loss_only is not None else args.prediction_loss_only

        # if eval is called w/o train init deepspeed here
        if args.deepspeed and not self.deepspeed:
            # XXX: eval doesn't have `resume_from_checkpoint` arg but we should be able to do eval
            # from the checkpoint eventually
            deepspeed_engine, _, _ = deepspeed_init(
                self, num_training_steps=0, resume_from_checkpoint=None, inference=True
            )
            self.model = deepspeed_engine.module
            self.model_wrapped = deepspeed_engine
            self.deepspeed = deepspeed_engine

        model = self._wrap_model(self.model, training=False, dataloader=dataloader)

        if not self.skip_eval_loss:
            # if full fp16 or bf16 eval is wanted and this ``evaluation`` or ``predict`` isn't called
            # while ``train`` is running, cast it to the right dtype first and then put on device
            if not self.is_in_train:
                if args.fp16_full_eval:
                    model = model.to(dtype=torch.float16, device=args.device)
                elif args.bf16_full_eval:
                    model = model.to(dtype=torch.bfloat16, device=args.device)

            batch_size = self.args.eval_batch_size

            logger.warning(f"***** Running {description} *****")
            if has_length(dataloader):
                logger.warning(f"  Num examples = {self.num_examples(dataloader)}")
            else:
                logger.info("  Num examples: Unknown")
            logger.info(f"  Batch size = {batch_size}")

            model.eval()

            self.callback_handler.eval_dataloader = dataloader
            # Do this before wrapping.
            eval_dataset = getattr(dataloader, "dataset", None)

            # MODIFIED: not necessary.
            if is_torch_tpu_available():
                dataloader = pl.MpDeviceLoader(dataloader, args.device)  # .per_device_loader(args.device)

            if args.past_index >= 0:
                self._past = None

            # Initialize containers
            # losses/preds/labels on GPU/TPU (accumulated for eval_accumulation_steps)
            losses_host = None
            preds_host = None
            labels_host = None
            inputs_host = None

            # losses/preds/labels on CPU (final containers)
            all_losses = None
            all_preds = None
            all_labels = None
            all_inputs = None
            # Will be useful when we have an iterable dataset so don't know its length.

            observed_num_examples = 0
            # Main evaluation loop
            for step, inputs in enumerate(dataloader):
                # Update the observed num examples
                observed_batch_size = find_batch_size(inputs)
                if observed_batch_size is not None:
                    observed_num_examples += observed_batch_size
                    # For batch samplers, batch_size is not known by the dataloader in advance.
                    if batch_size is None:
                        batch_size = observed_batch_size

                # Prediction step
                loss, logits, labels = self.prediction_step(
                    model, inputs, prediction_loss_only, ignore_keys=ignore_keys
                )
                inputs_decode = self._prepare_input(inputs["input_ids"]) if args.include_inputs_for_metrics else None

                # MODIFIED: not necessary.
                if is_torch_tpu_available():
                    xm.mark_step()

                # Update containers on host
                if loss is not None:
                    # MODIFIED: do not gather across devices. (different loss on each device)
                    losses = loss.repeat(batch_size)
                    losses_host = losses if losses_host is None else torch.cat((losses_host, losses), dim=0)
                if labels is not None:
                    labels = self._pad_across_processes(labels)
                    # MODIFIED: do not gather across devices.
                    labels = self._nested_gather(labels)
                    labels_host = (
                        labels if labels_host is None else nested_concat(labels_host, labels, padding_index=-100)
                    )
                if inputs_decode is not None:
                    inputs_decode = self._pad_across_processes(inputs_decode)
                    # MODIFIED: do not gather across devices.
                    inputs_decode = self._nested_gather(inputs_decode)
                    inputs_host = (
                        inputs_decode
                        if inputs_host is None
                        else nested_concat(inputs_host, inputs_decode, padding_index=-100)
                    )
                if logits is not None:
                    logits = self._pad_across_processes(logits)
                    # logits = self._nested_gather(logits)
                    if self.preprocess_logits_for_metrics is not None:
                        logits = self.preprocess_logits_for_metrics(logits, labels)
                    preds_host = logits if preds_host is None else nested_concat(preds_host, logits, padding_index=-100)
                self.control = self.callback_handler.on_prediction_step(args, self.state, self.control)

                # Gather all tensors and put them back on the CPU if we have done enough accumulation steps.
                if args.eval_accumulation_steps is not None and (step + 1) % args.eval_accumulation_steps == 0:
                    if losses_host is not None:
                        losses = nested_numpify(losses_host)
                        all_losses = losses if all_losses is None else np.concatenate((all_losses, losses), axis=0)
                    if preds_host is not None:
                        logits = nested_numpify(preds_host)
                        all_preds = (
                            logits if all_preds is None else nested_concat(all_preds, logits, padding_index=-100)
                        )
                    if inputs_host is not None:
                        inputs_decode = nested_numpify(inputs_host)
                        all_inputs = (
                            inputs_decode
                            if all_inputs is None
                            else nested_concat(all_inputs, inputs_decode, padding_index=-100)
                        )
                    if labels_host is not None:
                        labels = nested_numpify(labels_host)
                        all_labels = (
                            labels if all_labels is None else nested_concat(all_labels, labels, padding_index=-100)
                        )

                    # Set back to None to begin a new accumulation
                    losses_host, preds_host, inputs_host, labels_host = (
                        None,
                        None,
                        None,
                        None,
                    )

            if args.past_index and hasattr(self, "_past"):
                # Clean the state at the end of the evaluation loop
                delattr(self, "_past")

            # Gather all remaining tensors and put them back on the CPU
            if losses_host is not None:
                losses = nested_numpify(losses_host)
                all_losses = losses if all_losses is None else np.concatenate((all_losses, losses), axis=0)
            if preds_host is not None:
                logits = nested_numpify(preds_host)
                all_preds = logits if all_preds is None else nested_concat(all_preds, logits, padding_index=-100)
            if inputs_host is not None:
                inputs_decode = nested_numpify(inputs_host)
                all_inputs = (
                    inputs_decode
                    if all_inputs is None
                    else nested_concat(all_inputs, inputs_decode, padding_index=-100)
                )
            if labels_host is not None:
                labels = nested_numpify(labels_host)
                all_labels = labels if all_labels is None else nested_concat(all_labels, labels, padding_index=-100)

            # Number of samples
            if has_length(eval_dataset):
                num_samples = len(eval_dataset)
            # The instance check is weird and does not actually check for the type, but whether the dataset has the right
            # methods. Therefore we need to make sure it also has the attribute.
            elif isinstance(eval_dataset, IterableDatasetShard) and hasattr(eval_dataset, "num_examples"):
                num_samples = eval_dataset.num_examples
            else:
                if has_length(dataloader):
                    num_samples = self.num_examples(dataloader)
                else:  # both len(dataloader.dataset) and len(dataloader) fail
                    num_samples = observed_num_examples

            # Number of losses has been rounded to a multiple of batch_size and in a distributed training, the number of
            # samplers has been rounded to a multiple of batch_size, so we truncate.
            if all_losses is not None:
                all_losses = all_losses[:num_samples]
            if all_preds is not None:
                all_preds = nested_truncate(all_preds, num_samples)
            if all_labels is not None:
                all_labels = nested_truncate(all_labels, num_samples)
            if all_inputs is not None:
                all_inputs = nested_truncate(all_inputs, num_samples)
        else:
            if is_torch_tpu_available():
                xm.rendezvous("eval_metrics")
            all_losses, all_preds, all_labels, all_inputs, num_samples = None, None, None, None, 0

        # Metrics!
        if is_torch_tpu_available():
            xm.rendezvous("eval_metrics")
        # MODIFIED: always compute metrics
        if self.compute_metrics is not None:
            metrics = self.compute_metrics(self)
        else:
            metrics = {}

        # To be JSON-serializable, we need to remove numpy types or zero-d tensors
        metrics = denumpify_detensorize(metrics)

        if all_losses is not None:
            # MODIFIED: no gather, add prefix
            loss = all_losses.mean().item()
            metrics[f"{metric_key_prefix}_{self.logging_prefix}loss"] = loss
            metrics[f"{metric_key_prefix}_loss"] = loss

        # Prefix all keys with metric_key_prefix + '_'
        for key in list(metrics.keys()):
            if not key.startswith(f"{metric_key_prefix}_"):
                metrics[f"{metric_key_prefix}_{key}"] = metrics.pop(key)

        return EvalLoopOutput(
            predictions=all_preds,
            label_ids=all_labels,
            metrics=metrics,
            num_samples=num_samples,
        )

    def _save_tpu(self, output_dir: Optional[str] = None):
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        logger.info(f"Saving model checkpoint to {output_dir}")

        # MODIFIED: also save on other devices
        if xm.is_master_ordinal():
            os.makedirs(output_dir, exist_ok=True)
            torch.save(self.args, os.path.join(output_dir, TRAINING_ARGS_NAME))

        # # Save a trained model and configuration using `save_pretrained()`.
        # # They can then be reloaded using `from_pretrained()`
        xm.rendezvous("saving_checkpoint")
        if isinstance(self.model, Model):
            actual_model = self.model.backbone
        else:
            actual_model = self.model
        if not isinstance(actual_model, PreTrainedModel):
            if isinstance(unwrap_model(actual_model), PreTrainedModel):
                unwrap_model(actual_model).save_pretrained(
                    output_dir,
                    is_main_process=self.args.should_save,
                    state_dict=actual_model.state_dict(),
                    save_function=xm.save,
                )
            else:
                logger.warning("Trainer.model is not a `PreTrainedModel`, only saving its state dict.")
                state_dict = actual_model.state_dict()
                xm.save(state_dict, os.path.join(output_dir, WEIGHTS_NAME))
        else:
            actual_model.save_pretrained(output_dir, is_main_process=self.args.should_save, save_function=xm.save)
        if self.tokenizer is not None and self.args.should_save:
            self.tokenizer.save_pretrained(output_dir)


class AdapterTrainerCallback(TrainerCallback):
    def __init__(self, trainer):
        super().__init__()
        self.trainer = trainer

    def on_train_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        model = kwargs.pop("model")
        model_frozen = getattr(model.backbone.base_model, "model_frozen", False)
        if not model_frozen:
            raise ValueError(
                "The pre-trained model weights are not frozen. For training adapters, please call the train_adapter()"
                " method"
            )

    def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        # apply adapter fusion weight regularization on the value matrix
        model = kwargs.pop("model")
        if self.trainer.train_adapter_fusion:
            fusion_reg_loss = model.backbone.base_model.get_fusion_regularization_loss()
            if fusion_reg_loss is not None:
                fusion_reg_loss.backward()
